import hashlib
import uuid
from datetime import datetime
from pathlib import Path


def get_iso8601_timestamp_now_unaware():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")


def get_file_hash(file_path: Path) -> str:
    if isinstance(file_path, str):
        file_path = Path(file_path)
    return hashlib.md5(file_path.read_bytes()).hexdigest()


def generate_run_id() -> str:
    # Generate a random run guid
    return str(uuid.uuid4())



from pathlib import Path

# we should make this a proper package that loads examples
# similar how they handle it in xarray:
# https://github.com/pydata/xarray/blob/main/xarray/tutorial.py
# Currently there are simple paths to the examples which are then easily loaded in the notebooks

_EXAMPLES = Path(__file__).parent

power_grid_model = _EXAMPLES / "power-grid-example.xlsx"
power_grid_containers = _EXAMPLES / "power-grid-containers.yaml"
power_grid_data_model = _EXAMPLES / "power-grid-model.yaml"
simple_example = _EXAMPLES / "sheet2cdf-transformation-rules.xlsx"
source_to_solution_mapping = _EXAMPLES / "source-to-solution-mapping-rules.xlsx"
nordic44 = _EXAMPLES / "Rules-Nordic44-to-TNT.xlsx"
nordic44_graphql = _EXAMPLES / "Rules-Nordic44-to-graphql.xlsx"
skos = _EXAMPLES / "skos-rules.xlsx"
wind_energy_ontology = _EXAMPLES / "wind-energy.owl"



import logging
from pathlib import Path
from typing import Any, cast

from fastapi import APIRouter, Response
from rdflib import Namespace

from cognite.neat.app.api.configuration import NEAT_APP
from cognite.neat.app.api.data_classes.rest import TransformationRulesUpdateRequest
from cognite.neat.rules import exporter, importer
from cognite.neat.rules.models._base import EntityTypes
from cognite.neat.rules.models.rules import Class, Classes, Metadata, Properties, Property, Rules
from cognite.neat.workflows.steps.data_contracts import RulesData
from cognite.neat.workflows.steps.lib.v1.rules_importer import ImportExcelToRules
from cognite.neat.workflows.utils import get_file_hash

router = APIRouter()


@router.get("/api/rules/list")
def get_rules_list():
    rules_dir = Path(NEAT_APP.config.rules_store_path)
    return {"result": [str(file.name) for file in rules_dir.glob("*.xlsx")]}


@router.get("/api/rules")
def get_rules(
    sheetname: str = "Properties",
    url: str | None = None,
    source_type: str | None = None,
    orient: str = "columns",
    workflow_name: str = "default",
    file_name: str | None = None,
    version: str | None = None,
) -> dict[str, Any]:
    if NEAT_APP.cdf_store is None or NEAT_APP.workflow_manager is None:
        return {"error": "NeatApp is not initialized"}
    if workflow_name != "undefined":
        workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
        if workflow is None:
            return {"error": f"Workflow {workflow_name} is not found"}
        workflow_definition = workflow.get_workflow_definition()
        if not file_name:
            for step in workflow_definition.steps:
                if step.method == ImportExcelToRules.__name__:
                    file_name = step.configs["file_name"]
                    version = step.configs["version"]
                    break
    if not file_name:
        return {"error": "File name is not provided"}
    rules_file = Path(file_name)
    if str(rules_file.parent) == ".":
        path = Path(NEAT_APP.config.rules_store_path) / rules_file
    else:
        path = Path(NEAT_APP.config.data_store_path) / rules_file

    src = "local"
    if url:
        path = Path(url)

    if path.exists() and not version:
        logging.info(f"Loading rules from {path}")
    elif path.exists() and version:
        hash_ = get_file_hash(path)
        if hash_ != version:
            NEAT_APP.cdf_store.load_rules_file_from_cdf(file_name, version)
            src = "cdf"
    else:
        NEAT_APP.cdf_store.load_rules_file_from_cdf(file_name, version)
        src = "cdf"

    error_text = ""
    properties = []
    classes = []
    try:
        rules = cast(Rules, importer.ExcelImporter(path).to_rules(return_report=False, skip_validation=False))

        properties = [
            {
                "class": value.class_id,
                "property": value.property_id,
                "property_description": value.description,
                "property_type": (
                    value.expected_value_type.versioned_id
                    if value.property_type == EntityTypes.object_property
                    else value.expected_value_type.suffix
                ),
                "cdf_resource_type": value.cdf_resource_type,
                "cdf_metadata_type": value.resource_type_property,
                "rule_type": value.rule_type,
                "rule": value.rule,
            }
            for value in rules.properties.values()
        ]

        classes = [
            {
                "class": value.class_id,
                "class_description": value.description,
                "cdf_resource_type": value.cdf_resource_type,
            }
            for value in rules.classes.values()
        ]

    except Exception as e:
        error_text = str(e)

    return {
        "properties": properties,
        "metadata": rules.metadata.model_dump(),
        "classes": classes,
        "file_name": path.name,
        "hash": get_file_hash(path),
        "error_text": error_text,
        "src": src,
    }


@router.get("/api/rules/from_file")
def get_original_rules_from_file(file_name: str):
    # """Endpoint for retrieving raw transformation from file"""
    path = Path(NEAT_APP.config.rules_store_path) / file_name
    rules = cast(Rules, importer.ExcelImporter(filepath=path).to_rules(return_report=False, skip_validation=False))
    return Response(content=rules.model_dump_json(), media_type="application/json")


@router.get("/api/rules/from_workflow")
def get_original_rules_from_workflow(workflow_name: str):
    """Endpoing for retrieving transformation from memmory"""
    workflow = NEAT_APP.workflow_manager.get_workflow(workflow_name)
    if workflow is None:
        return {"error": f"Workflow {workflow_name} is not found"}
    context = workflow.get_context()
    rules_data = context["RulesData"]
    if type(rules_data) != RulesData:
        return {"error": "RulesData is not found in workflow context"}

    return Response(content=rules_data.rules.model_dump_json(), media_type="application/json")


@router.post("/api/rules/model_and_transformations")
def upsert_rules(request: TransformationRulesUpdateRequest):
    """Endpoing for updating transformation rules via API . This endpoint is still experimental"""
    rules = request.rules_object
    rules["metadata"]["namespace"] = Namespace(rules["metadata"]["namespace"])
    metadata = Metadata(**rules["metadata"])
    classes = Classes()
    for class_, val in rules["classes"].items():
        classes[class_] = Class(**val)
    properties = Properties()

    for prop, val in rules["properties"].items():
        val["resource_type_property"] = []
        properties[prop] = Property(**val)

    prefixes: dict[str, Namespace] = {}
    for prefix, val in rules["prefixes"].items():
        prefixes[prefix] = Namespace(val)

    rules = Rules(metadata=metadata, classes=classes, properties=properties, prefixes=prefixes, instances=[])
    if request.output_format == "excel":
        rules_file = Path(request.file_name)
        if str(rules_file.parent) == ".":
            path = Path(NEAT_APP.config.rules_store_path) / rules_file
        else:
            path = Path(NEAT_APP.config.data_store_path) / rules_file

        exporter.ExcelExporter(rules=rules).export_to_file(path)
    return {"status": "ok"}



import logging
import sys
import time
from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import Literal, TypeAlias, cast

import pandas as pd
from prometheus_client import Gauge, Summary
from rdflib import Graph, Namespace, URIRef
from rdflib.query import Result, ResultRow

from cognite.neat.constants import DEFAULT_NAMESPACE, PREFIXES
from cognite.neat.graph.models import Triple
from cognite.neat.graph.stores._rdf_to_graph import rdf_file_to_graph
from cognite.neat.rules.models.rules import Rules

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

prom_qsm = Summary("store_query_time_summary", "Time spent processing queries", ["query"])
prom_sq = Gauge("store_single_query_time", "Time spent processing a single query", ["query"])

MIMETypes: TypeAlias = Literal[
    "application/rdf+xml", "text/turtle", "application/n-triple", "application/n-quads", "application/trig"
]


class NeatGraphStoreBase(ABC):
    """NeatGraphStore is a class that stores the graph and provides methods to read/write data it contains


    Args:
        graph : Instance of rdflib.Graph class for graph storage
        base_prefix : Used as a base prefix for graph namespace, allowing querying graph data using a shortform of a URI
        namespace : Namespace (aka URI) used to resolve any relative URI in the graph
        prefixes : Dictionary of additional prefixes used and bounded to the graph
    """

    rdf_store_type: str

    def __init__(
        self,
        graph: Graph | None = None,
        base_prefix: str = "",  # usually empty
        namespace: Namespace = DEFAULT_NAMESPACE,
        prefixes: dict = PREFIXES,
    ):
        self.graph = graph or Graph()
        self.base_prefix: str = base_prefix
        self.namespace: Namespace = namespace
        self.prefixes: dict[str, Namespace] = prefixes

        self.rdf_store_query_url: str | None = None
        self.rdf_store_update_url: str | None = None
        self.returnFormat: str | None = None
        self.df_cache: pd.DataFrame | None = None
        self.internal_storage_dir: Path | None = None
        self.graph_name: str | None = None
        self.internal_storage_dir_orig: Path | None = None
        self.storage_dirs_to_delete: list[Path] = []
        self.queries = _Queries(self)

    @classmethod
    def from_rules(cls, rules: Rules) -> Self:
        """
        Creates a new instance of NeatGraphStore from TransformationRules and runs the .init_graph() method on it.

        Args:
            rules: TransformationRules object containing information about the graph store.

        Returns:
            An instantiated instance of NeatGraphStore

        """
        if rules.metadata.namespace is None:
            namespace = DEFAULT_NAMESPACE
        else:
            namespace = rules.metadata.namespace
        store = cls(prefixes=rules.prefixes, namespace=namespace)
        store.init_graph(base_prefix=rules.metadata.prefix)
        return store

    @abstractmethod
    def _set_graph(self) -> None:
        raise NotImplementedError()

    def init_graph(
        self,
        rdf_store_query_url: str | None = None,
        rdf_store_update_url: str | None = None,
        graph_name: str | None = None,
        base_prefix: str | None = None,
        returnFormat: str = "csv",
        internal_storage_dir: Path | None = None,
    ):
        """Initializes the graph.

        Args:
            rdf_store_query_url : URL towards which SPARQL query is executed, by default None
            rdf_store_update_url : URL towards which SPARQL update is executed, by default None
            graph_name : Name of graph, by default None
            base_prefix : Base prefix for graph namespace to change if needed, by default None
            returnFormat : Transport format of graph data between, by default "csv"
            internal_storage_dir : Path to directory where internal storage is located,
                                   by default None (in-memory storage).

        !!! note "internal_storage_dir"
            Used only for Oxigraph
        """
        logging.info("Initializing NeatGraphStore")
        self.rdf_store_query_url = rdf_store_query_url
        self.rdf_store_update_url = rdf_store_update_url
        self.graph_name = graph_name
        self.returnFormat = returnFormat
        self.internal_storage_dir = Path(internal_storage_dir) if internal_storage_dir else None
        self.internal_storage_dir_orig = (
            self.internal_storage_dir if self.internal_storage_dir_orig is None else self.internal_storage_dir_orig
        )

        self._set_graph()

        if self.prefixes:
            for prefix, namespace in self.prefixes.items():
                logging.info("Adding prefix %s with namespace %s", prefix, namespace)
                self.graph.bind(prefix, namespace)

        if base_prefix:
            self.base_prefix = base_prefix

        self.graph.bind(self.base_prefix, self.namespace)
        logging.info("Adding prefix %s with namespace %s", self.base_prefix, self.namespace)
        logging.info("Graph initialized")

    def reinitialize_graph(self):
        """Reinitialize the graph."""
        self.init_graph(
            self.rdf_store_query_url,
            self.rdf_store_update_url,
            self.graph_name,
            self.base_prefix,
            self.returnFormat,
            self.internal_storage_dir,
        )

    def upsert_prefixes(self, prefixes: dict[str, Namespace]) -> None:
        """Adds prefixes to the graph store."""
        self.prefixes.update(prefixes)
        for prefix, namespace in prefixes.items():
            logging.info("Adding prefix %s with namespace %s", prefix, namespace)
            self.graph.bind(prefix, namespace)

    def close(self) -> None:
        """Closes the graph."""
        # Can be overridden in subclasses
        return None

    def restart(self) -> None:
        """Restarts the graph"""
        # Can be overridden in subclasses
        return None

    def import_from_file(
        self, graph_file: Path, mime_type: MIMETypes = "application/rdf+xml", add_base_iri: bool = True
    ) -> None:
        """Imports graph data from file.

        Args:
            graph_file : File path to file containing graph data, by default None
            mime_type : MIME type of graph data, by default "application/rdf+xml"
            add_base_iri : Add base IRI to graph, by default True
        """
        if add_base_iri:
            self.graph = rdf_file_to_graph(
                self.graph, graph_file, base_namespace=self.namespace, prefixes=self.prefixes
            )
        else:
            self.graph = rdf_file_to_graph(self.graph, graph_file, prefixes=self.prefixes)
        return None

    def get_graph(self) -> Graph:
        """Returns the graph."""
        return self.graph

    def set_graph(self, graph: Graph):
        """Sets the graph."""
        self.graph = graph

    def query(self, query: str) -> Result:
        """Returns the result of the query."""
        start_time = time.perf_counter()
        result = self.graph.query(query)
        stop_time = time.perf_counter()
        elapsed_time = stop_time - start_time
        prom_qsm.labels("query").observe(elapsed_time)
        prom_sq.labels("query").set(elapsed_time)
        return result

    def serialize(self, *args, **kwargs):
        """Serializes the graph."""
        return self.graph.serialize(*args, **kwargs)

    def query_delayed(self, query) -> Iterable[Triple]:
        """Returns the result of the query, but does not execute it immediately.

        The query is not executed until the result is iterated over.

        Args:
            query: SPARQL query to execute

        Returns:
            An iterable of triples

        """
        return _DelayedQuery(self.graph, query)

    @abstractmethod
    def drop(self) -> None:
        """Drops the graph."""
        raise NotImplementedError()

    def garbage_collector(self) -> None:
        """Garbage collection of the graph store."""
        # Can be overridden in subclasses
        return None

    def query_to_dataframe(
        self,
        query: str,
        column_mapping: dict | None = None,
        save_to_cache: bool = False,
        index_column: str = "instance",
    ) -> pd.DataFrame:
        """Returns the result of the query as a dataframe.

        Args:
            query: SPARQL query to execute
            column_mapping: Columns name mapping, by default None
            save_to_cache: Save result of query to cache, by default False
            index_column: Indexing column , by default "instance"

        Returns:
            Dataframe with result of query
        """

        if column_mapping is None:
            column_mapping = {0: "instance", 1: "property", 2: "value"}

        result = self.graph.query(query, DEBUG=False)
        df_cache = pd.DataFrame(list(result))
        df_cache.rename(columns=column_mapping, inplace=True)
        df_cache[index_column] = df_cache[index_column].apply(lambda x: str(x))
        if save_to_cache:
            self.df_cache = df_cache
        return df_cache

    def commit(self):
        """Commits the graph."""
        self.graph.commit()

    def get_df(self) -> pd.DataFrame:
        """Returns the cached dataframe."""
        if self.df_cache is None:
            raise ValueError("Cache is empty. Run query_to_dataframe() first with save_to_cache.")
        return self.df_cache

    def get_instance_properties_from_cache(self, instance_id: str) -> pd.DataFrame:
        """Returns the properties of an instance."""
        if self.df_cache is None:
            raise ValueError("Cache is empty. Run query_to_dataframe() first with save_to_cache.")
        return self.df_cache.loc[self.df_cache["instance"] == instance_id]

    def print_triples(self):
        """Prints the triples of the graph."""
        for subj, pred, obj in self.graph:
            logging.info(f"Triple: {subj} {pred} {obj}")

    def diagnostic_report(self):
        """Returns the dictionary representation graph diagnostic data ."""
        return {
            "rdf_store_type": self.rdf_store_type,
            "base_prefix": self.base_prefix,
            "namespace": self.namespace,
            "prefixes": self.prefixes,
            "internal_storage_dir": self.internal_storage_dir,
            "rdf_store_query_url": self.rdf_store_query_url,
            "rdf_store_update_url": self.rdf_store_update_url,
        }

    def add_triples(self, triples: list[Triple] | set[Triple], batch_size: int = 10_000, verbose: bool = False):
        """Adds triples to the graph store in batches.

        Args:
            triples: list of triples to be added to the graph store
            batch_size: Batch size of triples per commit, by default 10_000
            verbose: Verbose mode, by default False
        """

        commit_counter = 0
        if verbose:
            logging.info(f"Committing total of {len(triples)} triples to knowledge graph!")
        total_number_of_triples = len(triples)
        number_of_uploaded_triples = 0

        def check_commit(force_commit: bool = False):
            """Commit nodes to the graph if batch counter is reached or if force_commit is True"""
            nonlocal commit_counter
            nonlocal number_of_uploaded_triples
            if force_commit:
                number_of_uploaded_triples += commit_counter
                self.graph.commit()
                if verbose:
                    logging.info(f"Committed {number_of_uploaded_triples} of {total_number_of_triples} triples")
                return
            commit_counter += 1
            if commit_counter >= batch_size:
                number_of_uploaded_triples += commit_counter
                self.graph.commit()
                if verbose:
                    logging.info(f"Committed {number_of_uploaded_triples} of {total_number_of_triples} triples")
                commit_counter = 0

        for triple in triples:
            self.graph.add(triple)
            check_commit()

        check_commit(force_commit=True)


class _DelayedQuery(Iterable):
    def __init__(self, graph_ref: Graph, query: str):
        self.graph_ref = graph_ref
        self.query = query

    def __iter__(self) -> Iterator[Triple]:
        start_time = time.perf_counter()
        result = self.graph_ref.query(self.query)
        stop_time = time.perf_counter()
        elapsed_time = stop_time - start_time
        prom_qsm.labels("query").observe(elapsed_time)
        prom_sq.labels("query").set(elapsed_time)
        return cast(Iterator[Triple], iter(result))


class _Queries:
    """Helper class for storing standard queries for the graph store."""

    def __init__(self, store: NeatGraphStoreBase):
        self.store = store

    def list_instances_ids_of_class(self, class_uri: URIRef, limit: int = -1) -> list[URIRef]:
        """Get instances ids for a given class

        Args:
            class_uri: Class for which instances are to be found
            limit: Max number of instances to return, by default -1 meaning all instances

        Returns:
            List of class instance URIs
        """
        query_statement = "SELECT DISTINCT ?subject WHERE { ?subject a <class> .} LIMIT X".replace(
            "class", class_uri
        ).replace("LIMIT X", "" if limit == -1 else f"LIMIT {limit}")
        return [cast(tuple, res)[0] for res in list(self.store.query(query_statement))]

    def list_instances_of_type(self, class_uri: URIRef) -> list[ResultRow]:
        """Get all triples for instances of a given class

        Args:
            class_uri: Class for which instances are to be found

        Returns:
            List of triples for instances of the given class
        """
        query = (
            f"SELECT ?instance ?prop ?value "
            f"WHERE {{ ?instance rdf:type <{class_uri}> . ?instance ?prop ?value . }} order by ?instance "
        )
        logging.info(query)
        # Select queries gives an iterable of result rows
        return cast(list[ResultRow], list(self.store.query(query)))



import logging
import sys
from collections.abc import Iterable, Mapping, Sequence
from dataclasses import dataclass, fields
from datetime import datetime
from typing import Any, Literal, TypeAlias, cast, overload
from warnings import warn

import numpy as np
import pandas as pd
from cognite.client import CogniteClient
from cognite.client.data_classes import Asset, AssetHierarchy, AssetList, AssetUpdate
from cognite.client.exceptions import CogniteDuplicatedError
from deepdiff import DeepDiff  # type: ignore[import]
from rdflib import Graph
from rdflib.term import URIRef

from cognite.neat.graph.loader.core.models import AssetTemplate
from cognite.neat.graph.stores import NeatGraphStoreBase
from cognite.neat.rules.models.rules import Property, Rules
from cognite.neat.utils.utils import chunker, datetime_utc_now, remove_namespace, retry_decorator

if sys.version_info >= (3, 11):
    from datetime import UTC
    from typing import Self
else:
    from datetime import timezone

    from typing_extensions import Self

    UTC = timezone.utc

EXCLUDE_PATHS = [
    "root['labels']",
    "root['metadata']['create_time']",
    "root['metadata']['start_time']",
    "root['metadata']['update_time']",
    "root['metadata']['end_time']",
    "root['metadata']['resurrection_time']",  # need to account for assets that are brought back to life
]


@dataclass
class NeatMetadataKeys:
    """Class holding mapping between NEAT metadata key names and their desired names in
    in CDF Asset metadata

    Args:
        start_time: Start time key name
        end_time: End time key name
        update_time: Update time key name
        resurrection_time: Resurrection time key name
        identifier: Identifier key name
        active: Active key name
        type: Type key name
    """

    start_time: str = "start_time"
    end_time: str = "end_time"
    update_time: str = "update_time"
    resurrection_time: str = "resurrection_time"
    identifier: str = "identifier"
    active: str = "active"
    type: str = "type"

    @classmethod
    def load(cls, data: dict) -> Self:
        cls_field_names = {f.name for f in fields(cls)}
        valid_keys = {}
        for key, value in data.items():
            if key in cls_field_names:
                valid_keys[key] = value
            else:
                logging.warning(f"Invalid key set {key}")

        return cls(**valid_keys)

    def as_aliases(self) -> dict[str, str]:
        return {str(field.default): getattr(self, field.name) for field in fields(self)}


def _get_class_instance_ids(graph: Graph, class_uri: URIRef, limit: int = -1) -> list[URIRef]:
    """Get instances ids for a given class

    Args:
        graph: Graph containing class instances
        class_uri: Class for which instances are to be found
        limit: Max number of instances to return, by default -1 meaning all instances

    Returns:
        List of class instance URIs
    """

    query_statement = "SELECT DISTINCT ?subject WHERE { ?subject a <class> .} LIMIT X".replace(
        "class", class_uri
    ).replace("LIMIT X", "" if limit == -1 else f"LIMIT {limit}")
    logging.debug(f"Query statement: {query_statement}")
    return [cast(tuple, res)[0] for res in list(graph.query(query_statement))]


def _get_class_instance(graph: Graph, instance: URIRef) -> list[tuple]:
    """Get instance by means of tuples containing property-value pairs
    Args:
        graph: Graph containing class instances
        instance: Instance URI

    Returns:
        list of property-value pairs for given instance
    """

    query_statement = "SELECT DISTINCT ?predicate ?object WHERE {<subject> ?predicate ?object .}".replace(
        "subject", instance
    )
    result = list(cast(tuple, graph.query(query_statement)))

    # Adds instance id for the sake of keep the chain of custody
    result += [(URIRef("http://purl.org/dc/terms/identifier"), instance)]

    return result


def _get_class_property_pairs(transformation_rules: Rules) -> dict[str, list[Property]]:
    """Define classes in terms of their properties

    Args:
    transformation_rules : Instance of TransformationRules containing class and property definitions

    Returns:
        Dict containing keys as class ids and list of their properties
    """

    classes: dict[str, list[Property]] = {}

    for property_ in transformation_rules.properties.keys():
        class_ = transformation_rules.properties[property_].class_id
        if class_ in classes:
            classes[class_] += [transformation_rules.properties[property_]]
        else:
            classes[class_] = [transformation_rules.properties[property_]]

    return classes


def _define_asset_class_mapping(transformation_rules: Rules) -> dict[str, dict[str, list]]:
    """Define mapping from class to asset properties

    Args:
        transformation_rules : Instance of TransformationRules containing class and property definitions

    Returns:
        Dict containing mapping from class to asset properties
    """
    solution2cdf_mapping_rules = _get_class_property_pairs(transformation_rules)

    asset_class_mapping: dict[str, dict[str, list]] = {}

    for class_, properties in solution2cdf_mapping_rules.items():
        asset_class_mapping[class_] = {
            "external_id": [],
            "name": [],
            "description": [],
            "parent_external_id": [],
            "metadata": [],
        }

        for property_ in properties:
            if "Asset" in property_.cdf_resource_type and property_.property_name != "*":
                for resource_type_property in property_.resource_type_property or []:
                    if (
                        resource_type_property in asset_class_mapping[class_]
                        and property_.property_name not in asset_class_mapping[class_][resource_type_property]
                    ):
                        asset_class_mapping[class_][resource_type_property] += [property_.property_name]

                    if property_.property_name not in asset_class_mapping[class_]["metadata"]:
                        # Todo; Why Nikola?  This adds for example name property to metadata? Isn't that
                        #   controlled by the resource_type_property? If you would like this behavior you
                        #   would set resource_type_property to ["metadata", "name"]?
                        asset_class_mapping[class_]["metadata"] += [property_.property_name]

    return asset_class_mapping


def _remap_class_properties(class_instance: dict, asset_class_mapping: dict) -> tuple[dict, set, set]:
    """Remaps original class instance properties to asset properties (e.g., external_id, name, description, metadata)

    Args:
        class_instance: Dictionary containing class instance properties and values
                        originating from RDF stripped from namespaces
        asset_class_mapping: Property mapping from class to asset

    Returns:
        Remapped class instance, set of missing asset properties and set of missing asset metadata
    """
    # Make distinction between missing properties that map into Asset fields
    # and missing RDF properties that are defined by sheet
    instance_properties = list(class_instance.keys())
    missing_properties = set()

    for property_group, ordered_properties in asset_class_mapping.items():
        if property_group != "metadata" and ordered_properties:
            if matching_property := next((a for a in ordered_properties if a in instance_properties), None):
                class_instance[property_group] = class_instance[matching_property]
            else:
                missing_properties.add(property_group)

    missing_metadata = set(asset_class_mapping["metadata"]).difference(set(instance_properties))

    return class_instance, missing_properties, missing_metadata


def _class2asset_instance(
    class_: str,
    class_instance: dict,
    asset_class_mapping: dict,
    data_set_id: int,
    meta_keys: NeatMetadataKeys,
    orphanage_asset_external_id: str | None = None,
    external_id_prefix: str | None = None,
    fallback_property: str = NeatMetadataKeys.identifier,
    empty_name_default: str = "Missing Name",
    add_missing_metadata: bool = True,
) -> dict[str, Any]:
    """Converts class instance to asset instance dictionary

    Args:
        class_: Class name which instance is being converted to asset instance
        class_instance: Dictionary containing class instance properties and values originating from RDF
                        stripped from namespaces
        asset_class_mapping: Property mapping from class to asset
        data_set_id: data set id to which asset belongs
        orphanage_asset_id: Orphanage asset external id, by default None
        external_id_prefix: External id prefix to be added to any external id, by default None
        fallback_property: Property from class instance to be used as fallback in case of
                           missing properties, by default "identifier"


    Returns:
        Asset instance dictionary
    """

    remapped_class_instance, missing_properties, missing_metadata = _remap_class_properties(
        class_instance, asset_class_mapping
    )

    # setting class instance type to class name
    remapped_class_instance[meta_keys.type] = class_
    # This will be a default case since we want to use original identifier as external_id
    # We are though dropping namespace from the original identifier (avoiding long-tail URIs)

    if "external_id" in missing_properties or asset_class_mapping["external_id"] == []:
        try:
            __extracted_from___class2asset_instance_49(
                remapped_class_instance, fallback_property, "external_id", class_
            )
        except Exception:
            __extracted_from___class2asset_instance_56(fallback_property, class_, remapped_class_instance)
    # This should not be the use case however to still have name of the object we are using
    # fallback property here as well (typically identifier)
    if "name" in missing_properties:
        try:
            __extracted_from___class2asset_instance_49(remapped_class_instance, fallback_property, "name", class_)
        except Exception:
            __extracted_from___class2asset_instance_56(fallback_property, class_, remapped_class_instance)

    # If object is expected to have parent, but parent is not provided, it is added to orphanage
    # This is typically sign of objects not following proposed ontology/data model/schema
    if "parent_external_id" in missing_properties and orphanage_asset_external_id:
        remapped_class_instance["parent_external_id"] = orphanage_asset_external_id

    if "name" in remapped_class_instance and remapped_class_instance["name"] == "":
        remapped_class_instance["name"] = empty_name_default
    # To maintain shape across of all assets of specific type we are adding missing metadata
    # keys as empty strings, this was request by Statnett
    # Generally this is bad practice, but more of a workaround of their bad data
    if missing_metadata and add_missing_metadata:
        msg = f"Adding missing metadata keys with values set to empty string for {class_}"
        msg += f" instance <{remapped_class_instance['identifier']}>. "
        logging.debug(msg)
        for key in missing_metadata:
            if key not in remapped_class_instance.keys():
                remapped_class_instance[key] = ""
                logging.debug(f"\tKey {key} added to <{remapped_class_instance['identifier']}> metadata!")

    asset_instance = AssetTemplate(
        **remapped_class_instance, external_id_prefix=external_id_prefix, data_set_id=data_set_id
    )
    # Removing field external_id_prefix from asset instance dictionary as it is only
    # convenience field for external_id and parent_external_id update in AssetTemplate
    return asset_instance.model_dump(exclude={"external_id_prefix"})


# TODO Rename this here and in `__class2asset_instance`
def __extracted_from___class2asset_instance_49(remapped_class_instance, fallback_property, arg2, class_):
    remapped_class_instance[arg2] = remapped_class_instance[fallback_property]
    msg = f"Missing external_id for {class_} instance <{remapped_class_instance['identifier']}>. "
    msg += f"Using value <{remapped_class_instance[fallback_property]}> provided "
    msg += f"by property <{fallback_property}>!"

    logging.debug(msg)


# TODO Rename this here and in `__class2asset_instance`
def __extracted_from___class2asset_instance_56(fallback_property, class_, remapped_class_instance):
    msg = f"Fallback property <{fallback_property}> not found for {class_} "
    msg += f"instance <{remapped_class_instance['identifier']}>."
    logging.error(msg)
    raise ValueError(msg)


def _list2dict(class_instance: list) -> dict[str, Any]:
    """Converting list of class instance properties and values to dictionary

    Args:
        class_instance: Class instance properties and values originating from RDF as list of tuples

    Returns:
        Class instance properties and values as dictionary
    """

    class_instance_dict: dict[str, Any] = {}
    for property_value_pair in class_instance:
        property_ = remove_namespace(property_value_pair[0])

        # Remove namespace from URIRef values, otherwise convert Literal to string
        # ideally this should react upon property type provided in sheet
        # however Assets only support string values
        value = (
            remove_namespace(property_value_pair[1])
            if isinstance(property_value_pair[1], URIRef)
            else str(property_value_pair[1])
        )

        if property_ in class_instance_dict and value not in class_instance_dict[property_]:
            class_instance_dict[property_] = (
                class_instance_dict[property_] + [value]
                if isinstance(class_instance_dict[property_], list)
                else [class_instance_dict[property_], value]
            )
        else:
            class_instance_dict[property_] = value

    return class_instance_dict


def rdf2assets(
    graph_store: NeatGraphStoreBase,
    rules: Rules,
    data_set_id: int,
    stop_on_exception: bool = False,
    use_orphanage: bool = True,
    meta_keys: NeatMetadataKeys | None = None,
    asset_external_id_prefix: str | None = None,
) -> dict[str, dict[str, Any]]:
    """Creates assets from RDF graph

    Args:
        graph_store : Graph containing RDF data
        rules : Instance of TransformationRules class containing transformation rules
        data_set_id: data set id to which assets belong
        stop_on_exception : Whether to stop upon exception.
        use_orphanage : Whether to use an orphanage for assets without parent_external_id
        meta_keys : The names of neat metadat keys to use.

    Returns:
        Dictionary representations of assets by external id.
    """
    meta_keys = NeatMetadataKeys() if meta_keys is None else meta_keys
    if rules.metadata.namespace is None:
        raise ValueError("Namespace must be provided in transformation rules!")
    namespace = rules.metadata.namespace

    orphanage_asset_external_id = f"{asset_external_id_prefix or ''}orphanage-{data_set_id}"

    graph = graph_store.get_graph()
    # Step 1: Create rdf to asset property mapping
    logging.info("Generating rdf to asset property mapping")
    asset_class_mapping = _define_asset_class_mapping(rules)

    # Step 4: Get ids of classes
    logging.info("Get ids of instances of classes")
    assets: dict[str, dict[str, Any]] = {}
    class_ids = {class_: _get_class_instance_ids(graph, namespace[class_]) for class_ in asset_class_mapping}
    # Step 5: Create Assets based on class instances
    logging.info("Create Assets based on class instances")
    meta_keys_aliases = meta_keys.as_aliases()
    for class_ in asset_class_mapping:
        # TODO: Rename class_id to instance_id
        class_ns = namespace[class_]
        logging.debug(f"Processing class <{class_ns}> . Number of instances: {len(class_ids[class_])}")
        progress_counter = 0
        # loading all instances into cache
        try:
            query = (
                f"SELECT ?instance ?prop ?value "
                f"WHERE {{ ?instance rdf:type <{class_ns}> . ?instance ?prop ?value . }} order by ?instance "
            )
            logging.info(query)
            response_df = graph_store.query_to_dataframe(query)
        except Exception as e:
            logging.error(f"Error while loading instances of class <{class_ns}> into cache. Reason: {e}")
            if stop_on_exception:
                raise e
            continue

        grouped_df = response_df.groupby("instance")

        for instance_id, group_df in grouped_df:
            try:
                instance_property_values = group_df.filter(items=["property", "value"]).values.tolist()
                instance_property_values += [(URIRef("http://purl.org/dc/terms/identifier"), URIRef(str(instance_id)))]

                # this will strip namespace from property names and values
                class_instance = _list2dict(instance_property_values)

                # class instance is repaired and converted to asset dictionary
                asset = _class2asset_instance(
                    class_,
                    class_instance,
                    asset_class_mapping[class_],
                    data_set_id,
                    meta_keys,
                    orphanage_asset_external_id if use_orphanage else None,  # we need only base external id
                    asset_external_id_prefix or None,
                    fallback_property=meta_keys.identifier,
                )

                # adding labels and timestamps
                asset["labels"] = [asset["metadata"][meta_keys.type], "non-historic"]
                now = str(datetime.now(UTC))
                asset["metadata"][meta_keys.start_time] = now
                asset["metadata"][meta_keys.update_time] = now
                asset["metadata"] = {meta_keys_aliases.get(k, k): v for k, v in asset["metadata"].items()}

                # log every 10000 assets
                if progress_counter % 10000 == 0:
                    logging.info(" Next 10000 Assets processed")

                assets[asset["external_id"]] = asset
                progress_counter += 1
            except Exception as ValidationError:
                logging.error(
                    f"Skipping class <{class_}> instance <{remove_namespace(str(instance_id))}>, "
                    f"reason:\n{ValidationError}\n"
                )
                if stop_on_exception:
                    raise ValidationError

        logging.debug(f"Class <{class_}> processed")

    if orphanage_asset_external_id not in assets:
        logging.warning(f"Orphanage with external id {orphanage_asset_external_id} not found in asset hierarchy!")
        logging.warning(f"Adding default orphanage with external id {orphanage_asset_external_id}")
        assets[orphanage_asset_external_id] = _create_orphanage(orphanage_asset_external_id, data_set_id, meta_keys)

    logging.info("Assets dictionary created")

    return assets


def rdf2asset_dictionary(
    graph_store: NeatGraphStoreBase,
    transformation_rules: Rules,
    stop_on_exception: bool = False,
    use_orphanage: bool = True,
) -> dict[str, dict[str, Any]]:
    warn("'rdf2asset_dictionary' is deprecated, please use 'rdf2assets' instead!", stacklevel=2)
    logging.warning("'rdf2asset_dictionary' is deprecated, please use 'rdf2assets' instead!")
    return rdf2assets(graph_store, transformation_rules, stop_on_exception, use_orphanage)


def _create_orphanage(orphanage_external_id: str, dataset_id: int, meta_keys: NeatMetadataKeys) -> dict:
    now = str(datetime_utc_now())
    return {
        "external_id": orphanage_external_id,
        "name": "Orphanage",
        "parent_external_id": None,
        "description": "Used to store all assets which parent does not exist",
        "metadata": {
            meta_keys.type: "Orphanage",
            "cdfResourceType": "Asset",
            meta_keys.identifier: "orphanage",
            meta_keys.active: "true",
            meta_keys.start_time: now,
            meta_keys.update_time: now,
        },
        "data_set_id": dataset_id,
        "labels": ["Orphanage", "non-historic"],
    }


def _asset2dict(asset: Asset) -> dict:
    """Return asset as dict representation

    Args:
        asset : Instance of Asset class

    Returns:
        Asset in dict representation
    """

    return {
        "external_id": asset.external_id,
        "name": asset.name,
        "description": asset.description,
        "parent_external_id": asset.parent_external_id,
        "data_set_id": asset.data_set_id,
        "metadata": asset.metadata,
    }


def _flatten_labels(labels: list[dict[str, str]]) -> set[str]:
    """Flatten labels"""
    result = set()
    if labels is None:
        return result
    for label in labels:
        if "externalId" in label:
            result.add(label["externalId"])
        elif "external_id" in label:
            result.add(label["external_id"])
        else:
            logging.warning(f"Label {label} does not have externalId")
    return result


def _is_historic(labels) -> bool:
    """Check if asset is historic"""
    return "historic" in labels


def _categorize_cdf_assets(
    client: CogniteClient, data_set_id: int, partitions: int
) -> tuple[pd.DataFrame | None, dict[str, set]]:
    """Categorize CDF assets

    Args:
        client : Instance of CogniteClient
        data_set_id : Id of data set
        partitions : Number of partitions

    Returns:
        CDF assets as pandas dataframe and dictionary with categorized assets
    """
    cdf_assets = client.assets.list(data_set_ids=data_set_id, limit=-1, partitions=partitions)

    cdf_assets = remove_non_existing_labels(client, cdf_assets)

    cdf_asset_df = AssetList(resources=cdf_assets).to_pandas()

    logging.info(f"Number of assets in CDF {len(cdf_asset_df)} that have been fetched")

    if cdf_asset_df.empty:
        return None, {"non-historic": set(), "historic": set()}
    if "labels" not in cdf_asset_df:
        # Add empty list for labels column.
        cdf_asset_df["labels"] = np.empty((len(cdf_asset_df), 0)).tolist()

    cdf_columns = set(cdf_asset_df.columns)
    expected_columns = {"external_id", "labels", "parent_external_id", "data_set_id", "name", "description", "metadata"}

    cdf_asset_df = cdf_asset_df[list(expected_columns.intersection(cdf_columns))]
    cdf_asset_df = cdf_asset_df.where(pd.notnull(cdf_asset_df), None)
    cdf_asset_df["labels"] = cdf_asset_df["labels"].apply(_flatten_labels).values  # type: ignore
    cdf_asset_df["is_historic"] = cdf_asset_df.labels.apply(_is_historic).values

    categorized_asset_ids = {
        "historic": set(cdf_asset_df[cdf_asset_df.is_historic].external_id.values),
        "non-historic": set(cdf_asset_df[~cdf_asset_df.is_historic].external_id.values),
    }

    cdf_asset_df.drop(["is_historic"], axis=1, inplace=True)
    msg = f"CDF assets categorized into {len(categorized_asset_ids['historic'])} historic"
    msg += f" and {len(categorized_asset_ids['non-historic'])} non-historic assets"
    logging.info(msg)

    return cdf_asset_df, categorized_asset_ids


def order_assets(assets: dict[str, dict]) -> list[Asset]:
    """Order assets in a way that parent assets are created before child assets

    Args:
    assets : List of assets to be created

    Returns:
        Ordered list of assets
    """
    hierarchy = AssetHierarchy([Asset(**asset) for asset in assets.values()], ignore_orphans=True)
    insert_dct = hierarchy.groupby_parent_xid()
    subtree_count = hierarchy.count_subtree(insert_dct)

    hierarchy = None

    asset_creation_order = pd.DataFrame.from_dict(subtree_count, orient="index", columns=["order"]).sort_values(
        by="order", ascending=False
    )
    asset_creation_order["external_id"] = asset_creation_order.index

    hierarchy = AssetList([Asset(**asset) for asset in assets.values()]).to_pandas()
    hierarchy = hierarchy.where(pd.notnull(hierarchy), None)
    hierarchy = hierarchy.merge(asset_creation_order, left_on="external_id", right_on="external_id")
    hierarchy = hierarchy.sort_values(by="order", ascending=False)
    hierarchy.reset_index(drop=True, inplace=True)
    hierarchy.labels = hierarchy.labels.apply(_flatten_labels)
    hierarchy.drop(["order"], axis=1, inplace=True)

    return [Asset(**row.to_dict()) for _, row in hierarchy.iterrows()]


def _assets_to_create(rdf_assets: dict, asset_ids: set) -> list[Asset]:
    """Return list of assets to be created

    Args:
        rdf_assets : Dictionary containing assets derived from knowledge graph (RDF)
        asset_ids : Set of asset ids to be created

    Returns:
        Ordered list of assets to be created
    """
    start_time = datetime_utc_now()
    if asset_ids:
        logging.info("Wrangling assets to be created into their final form")
        ordered_assets = order_assets({external_id: rdf_assets[external_id] for external_id in asset_ids})

        logging.info(f"Wrangling completed in {(datetime_utc_now() - start_time).seconds} seconds")
        return ordered_assets
    return []


def _assets_to_update(
    rdf_assets: dict,
    cdf_assets: pd.DataFrame | None,
    asset_ids: set,
    meta_keys: NeatMetadataKeys,
    exclude_paths: list = EXCLUDE_PATHS,
) -> tuple[list[Asset], dict[str, dict]]:
    """Return list of assets to be updated

    Args:
        rdf_assets : Dictionary containing assets derived from knowledge graph (RDF)
        cdf_assets : Dataframe containing assets from CDF
        asset_ids : Candidate assets to be updated
        meta_keys : The neat meta data keys.
        exclude_paths : Paths not to be checked when diffing rdf and cdf assets, by default EXCLUDE_PATHS

    Returns:
        List of assets to be updated and detailed report of changes per asset
    """

    start_time = datetime_utc_now()
    assets = []
    report = {}
    if not asset_ids:
        return [], {}
    logging.info("Wrangling assets to be updated into their final form")
    if cdf_assets is None:
        cdf_asset_subset = {}
    else:
        cdf_asset_subset = {
            row["external_id"]: row
            for row in cdf_assets[cdf_assets["external_id"].isin(asset_ids)].to_dict(orient="records")
        }
    for external_id in asset_ids:
        cdf_asset = cdf_asset_subset[external_id]
        diffing_result = DeepDiff(cdf_asset, rdf_assets[external_id], exclude_paths=exclude_paths)

        if diffing_result and f"root['metadata']['{meta_keys.active}']" not in diffing_result.affected_paths:
            asset = Asset(**rdf_assets[external_id])
            if asset.metadata is None:
                asset.metadata = {}
            try:
                asset.metadata[meta_keys.start_time] = cdf_asset[external_id]["metadata"][meta_keys.start_time]
            except KeyError:
                asset.metadata[meta_keys.start_time] = str(datetime.now(UTC))
            asset.metadata[meta_keys.update_time] = str(datetime.now(UTC))
            assets.append(asset)

            report[external_id] = dict(diffing_result)

    logging.info(f"Wrangling of {len(assets)} completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return assets, report


def _assets_to_resurrect(
    rdf_assets: dict, cdf_assets: pd.DataFrame | None, asset_ids: set, meta_keys: NeatMetadataKeys
) -> list[Asset]:
    """Returns list of assets to be resurrected

    Args:
        rdf_assets : Dictionary containing assets derived from knowledge graph (RDF)
        cdf_assets : Dataframe containing assets from CDF
        asset_ids : Set of asset ids to be resurrected

    Returns:
        List of assets to be resurrected
    """
    start_time = datetime_utc_now()
    assets = []
    if not asset_ids:
        return []
    logging.info("Wrangling assets to be resurrected into their final form")
    if cdf_assets is None:
        cdf_asset_subset = {}
    else:
        cdf_asset_subset = {
            row["external_id"]: row
            for row in cdf_assets[cdf_assets["external_id"].isin(asset_ids)].to_dict(orient="records")
        }
    for external_id in asset_ids:
        cdf_asset = cdf_asset_subset[external_id]

        asset = Asset(**rdf_assets[external_id])
        if asset.metadata is None:
            asset.metadata = {}
        now = str(datetime.now(UTC))
        try:
            asset.metadata[meta_keys.start_time] = cdf_asset[external_id]["metadata"][meta_keys.start_time]
        except KeyError:
            asset.metadata[meta_keys.start_time] = now
        asset.metadata[meta_keys.update_time] = now
        asset.metadata[meta_keys.resurrection_time] = now
        assets.append(asset)

    logging.info(f"Wrangling of {len(assets)} completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return assets


def _assets_to_decommission(
    cdf_assets: pd.DataFrame | None, asset_ids: set[str], meta_keys: NeatMetadataKeys
) -> list[Asset]:
    start_time = datetime_utc_now()

    assets = []
    if not asset_ids:
        return []
    logging.info("Wrangling assets to be decommissioned into their final form")
    if cdf_assets is None:
        cdf_asset_subset: dict[str, dict] = {}
    else:
        cdf_asset_subset = {
            row["external_id"]: row
            for row in cdf_assets[cdf_assets["external_id"].isin(asset_ids)].to_dict(orient="records")
        }

    for external_id in asset_ids:
        cdf_asset = cdf_asset_subset[external_id]

        now = str(datetime.now(UTC))
        cdf_asset["metadata"][meta_keys.update_time] = now
        cdf_asset["metadata"].pop(meta_keys.resurrection_time, None)
        cdf_asset["metadata"][meta_keys.end_time] = now
        cdf_asset["metadata"][meta_keys.active] = "false"
        try:
            cdf_asset["labels"].remove("non-historic")
        except KeyError:
            logging.info(f"Asset {external_id} missed label 'non-historic'")
        cdf_asset["labels"].add("historic")

        assets.append(Asset(**cdf_asset))

    logging.info(f"Wrangling completed in {(datetime_utc_now() - start_time).seconds} seconds")
    return assets


@overload
def categorize_assets(
    client: CogniteClient,
    rdf_assets: dict,
    data_set_id: int,
    return_report: Literal[False] = False,
    partitions: int = 2,
    stop_on_exception: bool = False,
    meta_keys: NeatMetadataKeys | None = None,
) -> dict:
    ...


@overload
def categorize_assets(
    client: CogniteClient,
    rdf_assets: dict,
    data_set_id: int,
    return_report: Literal[True],
    partitions: int = 2,
    stop_on_exception: bool = False,
    meta_keys: NeatMetadataKeys | None = None,
) -> tuple[dict, dict]:
    ...


def categorize_assets(
    client: CogniteClient,
    rdf_assets: dict,
    data_set_id: int,
    return_report: bool = False,
    partitions: int = 2,
    stop_on_exception: bool = False,
    meta_keys: NeatMetadataKeys | None = None,
) -> tuple[dict, dict] | dict:
    """Categorize assets on those that are to be created, updated and decommissioned

    Args:
        client : Instance of CogniteClient
        rdf_assets : Dictionary containing asset external_id - asset pairs
        data_set_id : Dataset id to which assets are to be/are stored
        partitions : Number of partitions to use when fetching assets from CDF, by default 2
        stop_on_exception : Whether to stop on exception or not, by default False
        return_report : Whether to report on the diffing results or not, by default False
        meta_keys : The metadata keys used by neat.

    Returns:
        dictionary containing asset category - list of asset pairs
    """
    meta_keys = NeatMetadataKeys() if meta_keys is None else meta_keys

    # TODO: Cache categorized assets somewhere instead of creating them
    cdf_assets, categorized_asset_ids = _categorize_cdf_assets(client, data_set_id, partitions)

    rdf_asset_ids = set(rdf_assets.keys())

    # ids to create
    create_ids = rdf_asset_ids.difference(
        categorized_asset_ids["historic"].union(categorized_asset_ids["non-historic"])
    )

    # ids potentially to update
    update_ids = rdf_asset_ids.intersection(categorized_asset_ids["non-historic"])

    # ids to decommission
    decommission_ids = categorized_asset_ids["non-historic"].difference(rdf_asset_ids)

    # ids to resurrect
    resurrect_ids = categorized_asset_ids["historic"].intersection(rdf_asset_ids)

    logging.info(f"Number of assets to create: { len(create_ids)}")
    logging.info(f"Number of assets to potentially update: { len(update_ids)}")
    logging.info(f"Number of assets to decommission: { len(decommission_ids)}")
    logging.info(f"Number of assets to resurrect: { len(resurrect_ids)}")

    categorized_assets_update, report_update = _assets_to_update(
        rdf_assets, cdf_assets, update_ids, meta_keys=meta_keys
    )
    report = {
        "create": create_ids,
        "resurrect": resurrect_ids,
        "decommission": decommission_ids,
        "update": report_update,
    }
    categorized_assets = {
        "create": _assets_to_create(rdf_assets, create_ids),
        "update": categorized_assets_update,
        "resurrect": _assets_to_resurrect(rdf_assets, cdf_assets, resurrect_ids, meta_keys),
        "decommission": _assets_to_decommission(cdf_assets, decommission_ids, meta_keys),
    }

    return (categorized_assets, report) if return_report else categorized_assets


def _micro_batch_push(
    client: CogniteClient,
    assets: Sequence[Asset | AssetUpdate],
    batch_size: int = 1000,
    push_type: str = "update",
    message: str = "Updated",
    max_retries: int = 1,
    retry_delay: int = 5,
):
    """Updates assets in batches of 1000

    Args:
    client : CogniteClient
        Instance of CogniteClient
    assets : list
        List of assets to be created or updated
    batch_size : int, optional
        Size of batch, by default 1000
    push_type : str, optional
        Type of push, either "update" or "create", by default "update"
    message : str, optional
        Message to logged, by default "Updated"
    """
    total = len(assets)
    counter = 0
    if push_type not in ["update", "create"]:
        logging.info(f"push_type {push_type} not supported")
        raise ValueError(f"push_type {push_type} not supported")
    for batch in chunker(assets, batch_size):
        counter += len(batch)
        start_time = datetime_utc_now()

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="microbatch-assets")
        def upsert_assets(batch):
            if push_type == "update":
                client.assets.update(batch)
            elif push_type == "create":
                client.assets.create_hierarchy(batch)

        try:
            upsert_assets(batch)
        except CogniteDuplicatedError:
            # this is handling of very rare case when some assets might be lost . Normally this should not happen.
            # Last attempt to recover
            client.assets.create_hierarchy(batch, upsert=True)

        delta_time = (datetime_utc_now() - start_time).seconds

        msg = f"{message} {counter} of {total} assets, batch processing time: {delta_time:.2f} "
        msg += f"seconds ETC: {delta_time * (total - counter) / (60*batch_size) :.2f} minutes"
        logging.info(msg)


def upload_assets(
    client: CogniteClient,
    categorized_assets: Mapping[str, Sequence[Asset | AssetUpdate]],
    batch_size: int = 5000,
    max_retries: int = 1,
    retry_delay: int = 3,
):
    """Uploads categorized assets to CDF

    Args:
    client : CogniteClient
        Instance of CogniteClient
    categorized_assets : Dict[str, list]
        dictionary containing asset category - list of asset pairs
    batch_size : int, optional
        Size of batch, by default 5000

    !!! note "batch_size"
        If batch size is set to 1 or None, all assets will be pushed to CDF in one go.
    """
    if batch_size:
        logging.info(f"Uploading assets in batches of {batch_size}")
        if categorized_assets["create"]:
            _micro_batch_push(
                client,
                categorized_assets["create"],
                batch_size,
                push_type="create",
                message="Created",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

        if categorized_assets["update"]:
            _micro_batch_push(
                client,
                categorized_assets["update"],
                batch_size,
                message="Updated",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

        if categorized_assets["resurrect"]:
            _micro_batch_push(
                client,
                categorized_assets["resurrect"],
                batch_size,
                message="Resurrected",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

        if categorized_assets["decommission"]:
            _micro_batch_push(
                client,
                categorized_assets["decommission"],
                batch_size,
                message="Decommissioned",
                max_retries=max_retries,
                retry_delay=retry_delay,
            )

    else:
        logging.info("Batch size not set, pushing all assets to CDF in one go!")

        @retry_decorator(max_retries=max_retries, retry_delay=retry_delay, component_name="create-assets")
        def create_assets():
            if categorized_assets["create"]:
                try:
                    client.assets.create_hierarchy(categorized_assets["create"])
                except CogniteDuplicatedError:
                    client.assets.create_hierarchy(categorized_assets["create"], upsert=True)

            if categorized_assets["update"]:
                client.assets.create_hierarchy(categorized_assets["update"], upsert=True, upsert_mode="replace")

            if categorized_assets["resurrect"]:
                client.assets.create_hierarchy(categorized_assets["resurrect"], upsert=True, upsert_mode="replace")

            if categorized_assets["decommission"]:
                client.assets.create_hierarchy(categorized_assets["decommission"], upsert=True, upsert_mode="replace")

        create_assets()


AssetLike: TypeAlias = Asset | dict[str, Any]


@overload
def remove_non_existing_labels(client: CogniteClient, assets: Sequence[AssetLike]) -> Sequence[AssetLike]:
    ...


@overload
def remove_non_existing_labels(client: CogniteClient, assets: Mapping[str, AssetLike]) -> Mapping[str, AssetLike]:
    ...


def remove_non_existing_labels(
    client: CogniteClient, assets: Sequence[AssetLike] | Mapping[str, AssetLike]
) -> Sequence[AssetLike] | Mapping[str, AssetLike]:
    cdf_labels = client.labels.list(limit=-1)
    if not cdf_labels:
        # No labels, nothing to check.
        return assets

    available_labels = {label.external_id for label in cdf_labels}

    def clean_asset_labels(asset: Asset | dict[str, Any]) -> Asset | dict[str, Any]:
        if isinstance(asset, Asset):
            asset.labels = [label for label in (asset.labels or []) if label.external_id in available_labels] or None
        elif isinstance(asset, dict) and "labels" in asset:
            asset["labels"] = [label for label in asset["labels"] if label in available_labels]
        return asset

    if isinstance(assets, Sequence):
        return [clean_asset_labels(a) for a in assets]

    elif isinstance(assets, dict):
        return {external_id: clean_asset_labels(a) for external_id, a in assets.items()}

    raise ValueError(f"Invalid format for Assets={type(assets)}")


def unique_asset_labels(assets: Iterable[Asset | dict[str, Any]]) -> set[str]:
    labels: set[str] = set()
    for asset in assets:
        if isinstance(asset, Asset):
            labels |= {label.external_id for label in (asset.labels or []) if label.external_id}
        elif isinstance(asset, dict) and (asset_labels := asset.get("labels")):
            labels |= set(asset_labels)
        else:
            raise ValueError(f"Unsupported {type(asset)}")
    return labels



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

# TODO: if this module grows too big, split it into several files and place under ./converter directory

from pathlib import Path

import pandas as pd
from pydantic_core import ErrorDetails
from rdflib import DC, DCTERMS, OWL, RDF, RDFS, SKOS, Graph

from cognite.neat.rules.importer._base import BaseImporter
from cognite.neat.rules.models.raw_rules import RawRules
from cognite.neat.rules.models.rules import Rules
from cognite.neat.rules.models.tables import Tables
from cognite.neat.rules.models.value_types import XSD_VALUE_TYPE_MAPPINGS

from ._owl2classes import parse_owl_classes
from ._owl2metadata import parse_owl_metadata
from ._owl2properties import parse_owl_properties


class OWLImporter(BaseImporter):
    """Convert OWL ontology to tables/ transformation rules / Excel file.

        Args:
            owl_filepath: Path to OWL ontology

    !!! Note
        OWL Ontologies typically lacks some information that is required for making a complete
        data model. This means that the methods .to_rules() will typically fail. Instead, it is recommended
        that you use the .to_spreadsheet() method to generate an Excel file, and then manually add the missing
        information to the Excel file. The Excel file can then be converted to a `Rules` object.

        Alternatively, one can set the `make_compliant` parameter to True to allow neat to attempt to make
        the imported rules compliant by adding default values for missing information, attaching dangling
        properties to default containers based on the property type, etc. One has to be aware
        that NEAT will be opinionated about how to make the ontology compliant, and that the resulting
        rules may not be what you expect.

    """

    def __init__(self, owl_filepath: Path):
        self.owl_filepath = owl_filepath

    def to_tables(self, make_compliant: bool = False) -> dict[str, pd.DataFrame]:
        graph = Graph()
        try:
            graph.parse(self.owl_filepath)
        except Exception as e:
            raise Exception(f"Could not parse owl file: {e}") from e

        # bind key namespaces
        graph.bind("owl", OWL)
        graph.bind("rdf", RDF)
        graph.bind("rdfs", RDFS)
        graph.bind("dcterms", DCTERMS)
        graph.bind("dc", DC)
        graph.bind("skos", SKOS)

        tables: dict[str, pd.DataFrame] = {
            Tables.metadata: parse_owl_metadata(graph, make_compliant=make_compliant),
            Tables.classes: parse_owl_classes(graph, make_compliant=make_compliant),
            Tables.properties: parse_owl_properties(graph, make_compliant=make_compliant),
        }

        if make_compliant:
            tables = make_tables_compliant(tables)
        # add sorting of classes and properties prior exporting

        tables[Tables.classes] = tables[Tables.classes].sort_values(by=["Class"])
        tables[Tables.properties] = tables[Tables.properties].sort_values(by=["Class", "Property"])

        return tables

    def to_raw_rules(self, make_compliant: bool = False) -> RawRules:
        """Creates `RawRules` object from the data."""

        tables = self.to_tables(make_compliant=make_compliant)

        return RawRules.from_tables(tables=tables, importer_type=self.__class__.__name__)

    def to_rules(
        self,
        return_report: bool = False,
        skip_validation: bool = False,
        validators_to_skip: set[str] | None = None,
        make_compliant: bool = False,
    ) -> tuple[Rules | None, list[ErrorDetails] | None, list | None] | Rules:
        """
        Creates `Rules` object from the data.

        Args:
            return_report: To return validation report. Defaults to False.
            skip_validation: Bypasses Rules validation. Defaults to False.
            validators_to_skip: List of validators to skip. Defaults to None.
            make_compliant: Flag for generating compliant rules, by default False

        Returns:
            Instance of `Rules`, which can be validated, not validated based on
            `skip_validation` flag, or partially validated if `validators_to_skip` is set,
            and optional list of errors and warnings if
            `return_report` is set to True.

        !!! Note "Skip Validation
            `skip_validation` flag should be only used for purpose when `Rules` object
            is exported to an Excel file. Do not use this flag for any other purpose!
        """

        raw_rules = self.to_raw_rules(make_compliant=make_compliant)

        return raw_rules.to_rules(return_report, skip_validation, validators_to_skip)


def make_tables_compliant(tables: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
    tables = _add_missing_classes(tables)
    tables = _add_missing_value_types(tables)
    tables = _add_properties_to_dangling_classes(tables)
    tables = _add_entity_type_property(tables)

    return tables


def _add_missing_classes(tables: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
    """Add missing classes to containers.

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing classes added to containers
    """

    missing_classes = set(tables[Tables.properties].Class.to_list()) - set(tables[Tables.classes].Class.to_list())

    rows = []
    for class_ in missing_classes:
        rows += [
            {
                "Class": class_,
                "Name": None,
                "Description": None,
                "Parent Class": None,
                "Deprecated": False,
                "Deprecation Date": None,
                "Replaced By": None,
                "Source": None,
                "Source Entity Name": None,
                "Match Type": None,
                "Comment": (
                    "Added by NEAT. "
                    "This is a class that a domain of a property but was not defined in the ontology. "
                    "It is added by NEAT to make the ontology compliant with CDF."
                ),
            }
        ]

    if rows:
        tables[Tables.classes] = pd.concat(
            [tables[Tables.classes], pd.DataFrame(rows)],
            ignore_index=True,
        )

    return tables


def _add_missing_value_types(tables: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
    """Add properties to classes that do not have any properties defined to them

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing properties added to containers
    """

    xsd_types = set(XSD_VALUE_TYPE_MAPPINGS.keys())
    referred_types = set(tables[Tables.properties]["Type"].to_list())
    defined_classes = set(tables[Tables.classes]["Class"].to_list())

    rows = []
    for class_ in referred_types.difference(defined_classes).difference(xsd_types):
        rows += [
            {
                "Class": class_,
                "Name": None,
                "Description": None,
                "Parent Class": None,
                "Deprecated": False,
                "Deprecation Date": None,
                "Replaced By": None,
                "Source": None,
                "Source Entity Name": None,
                "Match Type": None,
                "Comment": (
                    "Added by NEAT. "
                    "This is a class that a domain of a property but was not defined in the ontology. "
                    "It is added by NEAT to make the ontology compliant with CDF."
                ),
            }
        ]

    if rows:
        tables[Tables.classes] = pd.concat(
            [tables[Tables.classes], pd.DataFrame(rows)],
            ignore_index=True,
        )

    return tables


def _add_properties_to_dangling_classes(
    tables: dict[str, pd.DataFrame], properties_to_add: list[str] | None = None
) -> dict[str, pd.DataFrame]:
    """Add properties to classes that do not have any properties defined to them

    Args:
        tables: imported tables from owl ontology

    Returns:
        Updated tables with missing properties added to containers
    """

    if properties_to_add is None:
        properties_to_add = ["label"]
    undefined_classes = set(tables[Tables.classes].Class.to_list()) - set(tables[Tables.properties].Class.to_list())

    rows = []
    for class_ in undefined_classes:
        for property_ in properties_to_add:
            rows += [
                {
                    "Class": class_,
                    "Property": property_,
                    "Name": property_,
                    "Description": None,
                    "Type": "string",
                    "Min Count": None,
                    "Max Count": 1,
                    "Deprecated": False,
                    "Deprecation Date": None,
                    "Replaced By": None,
                    "Source": None,
                    "Source Entity Name": None,
                    "Match Type": None,
                    "Comment": "Added by NEAT. Default property to make the ontology compliant with CDF.",
                }
            ]

    if rows:
        tables[Tables.properties] = pd.concat(
            [tables[Tables.properties], pd.DataFrame(rows)],
            ignore_index=True,
        )

    return tables


def _add_entity_type_property(tables: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
    missing_entity_type = set(
        tables[Tables.properties].groupby("Class").filter(lambda x: "entityType" not in x.Property.to_list()).Class
    )

    rows = []
    for class_ in missing_entity_type:
        rows += [
            {
                "Class": class_,
                "Property": "entityType",
                "Name": "entityType",
                "Description": None,
                "Type": "string",
                "Min Count": None,
                "Max Count": 1,
                "Deprecated": False,
                "Deprecation Date": None,
                "Replaced By": None,
                "Source": None,
                "Source Entity Name": None,
                "Match Type": None,
                "Comment": "Added by NEAT. Default property added to make the ontology compliant with CDF.",
            }
        ]

    if rows:
        tables[Tables.properties] = pd.concat(
            [tables[Tables.properties], pd.DataFrame(rows)],
            ignore_index=True,
        )
    return tables



import warnings
from collections import defaultdict

import pandas as pd

from cognite.neat.rules.models.rdfpath import TransformationRuleType
from cognite.neat.rules.models.rules import Property, Rules


def get_defined_classes(transformation_rules: Rules) -> set[str]:
    """Returns classes that have properties defined for them in the data model.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Set of classes that have been defined in the data model
    """
    return {property.class_id for property in transformation_rules.properties.values()}


def get_classes_with_properties(transformation_rules: Rules) -> dict[str, list[Property]]:
    """Returns classes that have been defined in the data model.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Dictionary of classes with a list of properties defined for them
    """

    class_property_pairs: dict[str, list[Property]] = {}

    for property_ in transformation_rules.properties.values():
        class_ = property_.class_id
        if class_ in class_property_pairs:
            class_property_pairs[class_] += [property_]
        else:
            class_property_pairs[class_] = [property_]

    return class_property_pairs


def to_class_property_pairs(transformation_rules: Rules, only_rdfpath: bool = False) -> dict[str, dict[str, Property]]:
    """Returns a dictionary of classes with a dictionary of properties associated with them.

    Args:
        transformation_rules : Instance of TransformationRules holding the data model
        only_rdfpath : To consider only properties which have rule `rdfpath` set. Defaults False

    Returns:
        Dictionary of classes with a dictionary of properties associated with them.

    !!! note "only_rdfpath"
        If only_rdfpath is True, only properties with RuleType.rdfpath will be returned as
        a part of the dictionary of properties related to a class. Otherwise, all properties
        will be returned.
    """

    class_property_pairs = {}

    for class_, properties in get_classes_with_properties(transformation_rules).items():
        processed_properties = {}
        for property_ in properties:
            if property_.property_id in processed_properties:
                # TODO: use appropriate Warning class from _exceptions.py
                # if missing make one !
                warnings.warn(
                    "Property has been defined more than once! Only first definition will be considered.", stacklevel=2
                )
                continue

            if (only_rdfpath and property_.rule_type == TransformationRuleType.rdfpath) or not only_rdfpath:
                processed_properties[property_.property_id] = property_
        class_property_pairs[class_] = processed_properties

    return class_property_pairs


def get_class_linkage(transformation_rules: Rules) -> pd.DataFrame:
    """Returns a dataframe with the class linkage of the data model.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Dataframe with the class linkage of the data model
    """

    class_linkage = pd.DataFrame(columns=["source_class", "target_class", "connecting_property", "max_occurrence"])
    for property_ in transformation_rules.properties.values():
        if property_.property_type == "ObjectProperty":
            new_row = pd.Series(
                {
                    "source_class": property_.class_id,
                    "target_class": property_.expected_value_type.suffix,
                    "connecting_property": property_.property_id,
                    "max_occurrence": property_.max_count,
                    "linking_type": "hierarchy" if property_.resource_type_property else "relationship",
                }
            )
            class_linkage = pd.concat([class_linkage, new_row.to_frame().T], ignore_index=True)

    class_linkage.drop_duplicates(inplace=True)

    return class_linkage


def get_class_hierarchy_linkage(rules: Rules) -> pd.DataFrame:
    """Remove linkage which is not creating asset hierarchy."""
    class_linkage = get_class_linkage(rules)
    return class_linkage[class_linkage.linking_type == "hierarchy"]


def get_connected_classes(transformation_rules: Rules) -> set[str]:
    """Return a set of classes that are connected to other classes.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Set of classes that are connected to other classes
    """
    class_linkage = get_class_linkage(transformation_rules)
    return set(class_linkage.source_class.values).union(set(class_linkage.target_class.values))


def get_disconnected_classes(transformation_rules: Rules) -> set[str]:
    """Return a set of classes that are disconnected (i.e. isolated) from other classes.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Set of classes that are disconnected from other classes
    """
    return get_defined_classes(transformation_rules) - get_connected_classes(transformation_rules)


def get_symmetric_pairs(transformation_rules: Rules) -> set[tuple[str, str]]:
    """Returns a set of pairs of symmetrically linked classes.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Set of pairs of symmetrically linked classes
    """

    # TODO: Find better name for this method
    sym_pairs: set[tuple[str, str]] = set()

    class_linkage = get_class_linkage(transformation_rules)
    if class_linkage.empty:
        return sym_pairs

    for _, row in class_linkage.iterrows():
        source = row.source_class
        target = row.target_class
        target_targets = class_linkage[class_linkage.source_class == target].target_class.values
        if source in target_targets and (source, target) not in sym_pairs:
            sym_pairs.add((source, target))
    return sym_pairs


def get_entity_ids(transformation_rules: Rules) -> set[str]:
    """Returns a set of entity ids (classes and properties) defined in the data model.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Set of entity ids (classes and properties) defined in the data model
    """
    return set(transformation_rules.classes.keys()).union(
        {property_.property_id for property_ in transformation_rules.properties.values()}
    )


def to_property_dict(transformation_rules: Rules) -> dict[str, list[Property]]:
    """Convert list of properties to a dictionary of lists of properties with property_id as key.

    Args:
        transformation_rules: Instance of TransformationRules holding the data model

    Returns:
        Dictionary of lists of properties with property_id as key
    """
    property_: dict[str, list[Property]] = defaultdict(list)

    for prop in transformation_rules.properties.values():
        if not (prop.property_id and prop.property_name == "*"):
            property_[prop.property_id].append(prop)

    return property_


def get_asset_related_properties(properties: list[Property]) -> list[Property]:
    """Return properties that are used to define CDF Assets

    Args:
        properties: List of properties

    Returns:
        List of properties that are used to define CDF Assets
    """
    return [prop for prop in properties if "Asset" in prop.cdf_resource_type]


def define_class_asset_mapping(transformation_rules: Rules, class_: str) -> dict[str, list[str]]:
    """Define mapping between class and asset properties

    Args:
        transformation_rules: Instance of TransformationRules holding the data model
        class_: Class id for which mapping is to be defined

    Returns:
        Dictionary with asset properties as keys and list of class properties as values
    """
    mapping_dict: dict[str, list[str]] = {}

    class_properties = to_class_property_pairs(transformation_rules, only_rdfpath=True)[class_]

    for asset_property in get_asset_related_properties(list(class_properties.values())):
        for resource_type_property in asset_property.resource_type_property or []:
            if resource_type_property not in mapping_dict:
                mapping_dict[resource_type_property] = [asset_property.property_id]
            else:
                mapping_dict[resource_type_property] += [asset_property.property_id]

    return mapping_dict



import hashlib
import logging
import sys
import time
from collections import OrderedDict
from collections.abc import Iterable
from datetime import datetime
from functools import wraps
from typing import TypeAlias, cast, overload

import pandas as pd
from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import CredentialProvider, OAuthClientCredentials, OAuthInteractive, Token
from cognite.client.exceptions import CogniteDuplicatedError, CogniteReadTimeout
from pydantic import HttpUrl, TypeAdapter, ValidationError
from pydantic_core import ErrorDetails
from pyparsing import Any
from rdflib import Literal, Namespace
from rdflib.term import URIRef

from cognite.neat import _version
from cognite.neat.utils.cdf import CogniteClientConfig, InteractiveCogniteClient, ServiceCogniteClient

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    from datetime import timezone

    UTC = timezone.utc


Triple: TypeAlias = tuple[URIRef, URIRef, Literal | URIRef]


def get_cognite_client_from_config(config: ServiceCogniteClient) -> CogniteClient:
    credentials = OAuthClientCredentials(
        token_url=config.token_url, client_id=config.client_id, client_secret=config.client_secret, scopes=config.scopes
    )

    return _get_cognite_client(config, credentials)


def get_cognite_client_from_token(config: ServiceCogniteClient) -> CogniteClient:
    credentials = Token(config.client_secret)
    return _get_cognite_client(config, credentials)


def get_cognite_client_interactive(config: InteractiveCogniteClient) -> CogniteClient:
    credentials = OAuthInteractive(
        authority_url=config.authority_url,
        client_id=config.client_id,
        scopes=config.scopes,
        redirect_port=config.redirect_port,
    )
    return _get_cognite_client(config, credentials)


def _get_cognite_client(config: CogniteClientConfig, credentials: CredentialProvider) -> CogniteClient:
    logging.info(f"Creating CogniteClient with parameters : {config}")

    # The client name is used for aggregated logging of Neat Usage
    client_name = f"CogniteNeat:{_version.__version__}"
    return CogniteClient(
        ClientConfig(
            client_name=client_name,
            base_url=config.base_url,
            project=config.project,
            credentials=credentials,
            timeout=config.timeout,
            max_workers=config.max_workers,
            debug=False,
        )
    )


@overload
def remove_namespace(*URI: URIRef | str, special_separator: str = "#_") -> str:
    ...


@overload
def remove_namespace(*URI: tuple[URIRef | str, ...], special_separator: str = "#_") -> tuple[str, ...]:
    ...


def remove_namespace(
    *URI: URIRef | str | tuple[URIRef | str, ...], special_separator: str = "#_"
) -> tuple[str, ...] | str:
    """Removes namespace from URI

    Args
        URI: URIRef | str
            URI of an entity
        special_separator : str
            Special separator to use instead of # or / if present in URI
            Set by default to "#_" which covers special client use case

    Returns
        Entities id without namespace

    Examples:

        >>> remove_namespace("http://www.example.org/index.html#section2")
        'section2'
        >>> remove_namespace("http://www.example.org/index.html#section2", "http://www.example.org/index.html#section3")
        ('section2', 'section3')
    """
    if isinstance(URI, str | URIRef):
        uris = (URI,)
    elif isinstance(URI, tuple):
        # Assume that all elements in the tuple are of the same type following type hint
        uris = cast(tuple[URIRef | str, ...], URI)
    else:
        raise TypeError(f"URI must be of type URIRef or str, got {type(URI)}")

    output = []
    for u in uris:
        try:
            _ = TypeAdapter(HttpUrl).validate_python(u)
            output.append(u.split(special_separator if special_separator in u else "#" if "#" in u else "/")[-1])
        except ValidationError:
            output.append(str(u))

    return tuple(output) if len(output) > 1 else output[0]


def get_namespace(URI: URIRef, special_separator: str = "#_") -> str:
    """Removes namespace from URI

    Parameters
    ----------
    URI : URIRef
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers special client use case

    Returns
    -------
    str
        Entity id without namespace
    """
    if special_separator in URI:
        return URI.split(special_separator)[0] + special_separator
    elif "#" in URI:
        return URI.split("#")[0] + "#"
    else:
        return "/".join(URI.split("/")[:-1]) + "/"


def convert_rdflib_content(content: Literal | URIRef | dict | list) -> Any:
    if isinstance(content, Literal) or isinstance(content, URIRef):
        return content.toPython()
    elif isinstance(content, dict):
        return {key: convert_rdflib_content(value) for key, value in content.items()}
    elif isinstance(content, list):
        return [convert_rdflib_content(item) for item in content]
    else:
        return content


def uri_to_short_form(URI: URIRef, prefixes: dict[str, Namespace]) -> str | URIRef:
    """Returns the short form of a URI if its namespace is present in the prefixes dict,
    otherwise returns the URI itself

    Args:
        URI: URI to be converted to form prefix:entityName
        prefixes: dict of prefixes

    Returns:
        shortest form of the URI if its namespace is present in the prefixes dict,
        otherwise returns the URI itself
    """
    uris: set[str | URIRef] = {URI}
    for prefix, namespace in prefixes.items():
        if URI.startswith(namespace):
            uris.add(f"{prefix}:{URI.replace(namespace, '')}")
    return min(uris, key=len)


def _traverse(hierarchy: dict, graph: dict, names: list[str]) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def get_generation_order(
    class_linkage: pd.DataFrame, parent_col: str = "source_class", child_col: str = "target_class"
) -> dict:
    parent_child_list = class_linkage[[parent_col, child_col]].values.tolist()
    # Build a directed graph and a list of all names that have no parent
    graph: dict[str, set[str]] = {name: set() for tup in parent_child_list for name in tup}
    has_parent = {name: False for tup in parent_child_list for name in tup}
    for parent, child in parent_child_list:
        graph[parent].add(child)
        has_parent[child] = True

    # All names that have absolutely no parent:
    roots = [name for name, parents in has_parent.items() if not parents]

    return _traverse({}, graph, roots)


def prettify_generation_order(generation_order: dict, depth: dict | None = None, start=-1) -> dict:
    """Prettifies generation order dictionary for easier consumption."""
    depth = depth or {}
    for key, value in generation_order.items():
        depth[key] = start + 1
        if isinstance(value, dict):
            prettify_generation_order(value, depth, start=start + 1)
    return OrderedDict(sorted(depth.items(), key=lambda item: item[1]))


def epoch_now_ms() -> int:
    return int((datetime.now(UTC) - datetime(1970, 1, 1, tzinfo=UTC)).total_seconds() * 1000)


def chunker(sequence, chunk_size):
    return [sequence[i : i + chunk_size] for i in range(0, len(sequence), chunk_size)]


def datetime_utc_now():
    return datetime.now(UTC)


def retry_decorator(max_retries=2, retry_delay=3, component_name=""):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            previous_exception = None
            for attempt in range(max_retries + 1):
                try:
                    logging.debug(f"Attempt {attempt + 1} of {max_retries + 1} for {component_name}")
                    return func(*args, **kwargs)
                except CogniteReadTimeout as e:
                    previous_exception = e
                    if attempt < max_retries:
                        logging.error(
                            f"""CogniteReadTimeout retry attempt {attempt + 1} failed for {component_name} .
                            Retrying in {retry_delay} second(s). Error:"""
                        )
                        logging.error(e)
                        time.sleep(retry_delay)
                    else:
                        raise e
                except CogniteDuplicatedError as e:
                    if isinstance(previous_exception, CogniteReadTimeout):
                        # if previous exception was CogniteReadTimeout,
                        # we can't be sure if the items were created or not
                        if len(e.successful) == 0 and len(e.failed) == 0 and len(e.duplicated) >= 0:
                            logging.warning(
                                f"Duplicate error for {component_name} . All items already exist in CDF. "
                                "Suppressing error."
                            )
                            return
                        else:
                            # can happend because of eventual consistency. Retry with delay to allow for CDF to catch up
                            if attempt < max_retries:
                                logging.error(
                                    f"""CogniteDuplicatedError retry attempt {attempt + 1} failed for {component_name} .
                                      Retrying in {retry_delay} second(s). Error:"""
                                )
                                logging.error(e)
                                # incerasing delay to allow for CDF to catch up
                                time.sleep(retry_delay)
                            else:
                                raise e
                    else:
                        # no point in retrying duplicate error if previous exception was not a timeout
                        raise e

                except Exception as e:
                    previous_exception = e
                    if attempt < max_retries:
                        logging.error(
                            f"Retry attempt {attempt + 1} failed for {component_name}. "
                            f"Retrying in {retry_delay} second(s)."
                        )
                        logging.error(e)
                        time.sleep(retry_delay)
                    else:
                        raise e

        return wrapper

    return decorator


def create_sha256_hash(string: str) -> str:
    # Create a SHA-256 hash object
    sha256_hash = hashlib.sha256()

    # Convert the string to bytes and update the hash object
    sha256_hash.update(string.encode("utf-8"))

    # Get the hexadecimal representation of the hash
    hash_value = sha256_hash.hexdigest()

    return hash_value


def generate_exception_report(exceptions: list[dict] | list[ErrorDetails] | None, category: str = "") -> str:
    exceptions_as_dict = _order_expectations_by_type(exceptions) if exceptions else {}
    report = ""

    for exception_type in exceptions_as_dict.keys():
        title = f"# {category}: {exception_type}" if category else ""
        warnings = "\n- " + "\n- ".join(exceptions_as_dict[exception_type])
        report += title + warnings + "\n\n"

    return report


def _order_expectations_by_type(exceptions: list[dict] | list[ErrorDetails]) -> dict[str, list[str]]:
    exception_dict: dict[str, list[str]] = {}
    for exception in exceptions:
        if not isinstance(exception["loc"], str) and isinstance(exception["loc"], Iterable):
            location = f"[{'/'.join(str(e) for e in exception['loc'])}]"
        else:
            location = ""

        issue = f"{exception['msg']} {location}"

        if exception_dict.get(exception["type"]) is None:
            exception_dict[exception["type"]] = [issue]
        else:
            exception_dict[exception["type"]].append(issue)
    return exception_dict


def remove_none_elements_from_set(s):
    return {x for x in s if x is not None}


def get_inheritance_path(child: Any, child_parent: dict[Any, list[Any]]) -> list:
    """Returns the inheritance path for a given child

    Args:
        child: Child class
        child_parent: Dictionary of child to parent relationship

    Returns:
        Inheritance path for a given child

    !!! note "No Circular Inheritance"
        This method assumes that the child_parent dictionary is a tree and does not contain any cycles.
    """
    path = []
    if child in child_parent:
        path.extend(child_parent[child])
        for parent in child_parent[child]:
            path.extend(get_inheritance_path(parent, child_parent))
    return path



"""This module is used to generate mock graph data for purposes of testing of NEAT.
"""
import logging
import random
import warnings
from collections import OrderedDict

import numpy
import pandas as pd
from prometheus_client import Gauge
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat.graph.models import Triple
from cognite.neat.rules.analysis import (
    get_class_linkage,
    get_classes_with_properties,
    get_defined_classes,
    get_symmetric_pairs,
)
from cognite.neat.rules.exporter._rules2rules import subset_rules
from cognite.neat.rules.models import Rules
from cognite.neat.rules.models.value_types import XSD_VALUE_TYPE_MAPPINGS
from cognite.neat.utils.utils import remove_namespace

from ._base import BaseExtractor

neat_total_processed_mock_triples = Gauge(
    "neat_total_processed_mock_triples", "Number of processed mock triples", ["state"]
)
neat_mock_triples_processing_timing = Gauge(
    "neat_mock_triples_processing_timing", "Generation of mock knowledge graph timing metrics", ["aggregation"]
)


class MockGraphGenerator(BaseExtractor):
    """
    Class used to generate mock graph data for purposes of testing of NEAT.

    Args:
        rules: Transformation rules defining the classes with their properties.
        class_count: Target class count for each class in the ontology
        stop_on_exception: To stop if exception is encountered or not, default is False
        allow_isolated_classes: To allow generation of instances for classes that are not
                                 connected to any other class, default is True
    """

    def __init__(
        self, rules: Rules, class_count: dict, stop_on_exception: bool = False, allow_isolated_classes: bool = True
    ):
        self.rules = rules
        self.class_count = class_count
        self.stop_on_exception = stop_on_exception
        self.allow_isolated_classes = allow_isolated_classes

    def extract(self) -> list[Triple]:
        """Generate mock triples based on data model defined transformation rules and desired number
        of class instances

        Returns:
            List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
        """
        return generate_triples(
            self.rules,
            self.class_count,
            stop_on_exception=self.stop_on_exception,
            allow_isolated_classes=self.allow_isolated_classes,
        )


def generate_triples(
    transformation_rules: Rules, class_count: dict, stop_on_exception: bool = False, allow_isolated_classes: bool = True
) -> list[Triple]:
    """Generate mock triples based on data model defined transformation rules and desired number
    of class instances

    Args:
        transformation_rules : Transformation rules defining the data model
        class_count: Target class count for each class in the ontology
        stop_on_exception: To stop if exception is encountered or not, default is False
        allow_isolated_classes: To allow generation of instances for classes that are not
                                 connected to any other class, default is True

    Returns:
        List of RDF triples, represented as tuples `(subject, predicate, object)`, that define data model instances
    """

    # Figure out which classes are defined in the data model and which are not
    if transformation_rules.metadata.namespace is None:
        raise ValueError("Namespace must be defined in transformation rules!")
    namespace = transformation_rules.metadata.namespace

    defined_classes = get_defined_classes(transformation_rules)

    if non_existing_classes := set(class_count.keys()) - defined_classes:
        msg = f"Class count contains classes {non_existing_classes} for which properties are not defined in Data Model!"
        if stop_on_exception:
            logging.error(msg)
            raise ValueError(msg)
        else:
            msg += " These classes will be ignored."
            logging.warning(msg)
            warnings.warn(msg, stacklevel=2)
            for class_ in non_existing_classes:
                class_count.pop(class_)

    # Subset data model to only classes that are defined in class count
    transformation_rules = (
        subset_rules(transformation_rules, set(class_count.keys()), skip_validation=True)
        if defined_classes != set(class_count.keys())
        else transformation_rules
    )

    class_linkage = get_class_linkage(transformation_rules)

    # Remove one of symmetric pairs from class linkage to maintain proper linking
    # among instances of symmetrically linked classes
    if sym_pairs := get_symmetric_pairs(transformation_rules):
        class_linkage = _remove_higher_occurring_sym_pair(class_linkage, sym_pairs)

    # Remove any of symmetric pairs containing classes that are not present class count
    class_linkage = _remove_non_requested_sym_pairs(class_linkage, class_count)

    # Generate generation order for classes instances
    generation_order = _prettify_generation_order(_get_generation_order(class_linkage))

    # Generated simple view of data model
    class_definitions = _rules_to_dict(transformation_rules)

    # pregenerate instance ids for each remaining class
    instance_ids = {key: [URIRef(namespace[f"{key}-{i}"]) for i in range(value)] for key, value in class_count.items()}

    # create triple for each class instance defining its type
    triples: list[Triple] = []
    for class_ in class_count:
        triples += [
            (class_instance_id, RDF.type, URIRef(namespace[class_])) for class_instance_id in instance_ids[class_]
        ]

    # generate triples for connected classes
    for class_ in generation_order:
        triples += _generate_triples_per_class(
            class_, class_definitions, sym_pairs, instance_ids, namespace, stop_on_exception
        )

    # generate triples for isolated classes
    if allow_isolated_classes:
        for class_ in set(class_count.keys()) - set(generation_order):
            triples += _generate_triples_per_class(
                class_, class_definitions, sym_pairs, instance_ids, namespace, stop_on_exception
            )

    return triples


def _get_generation_order(
    class_linkage: pd.DataFrame, parent_col: str = "source_class", child_col: str = "target_class"
) -> dict:
    parent_child_list: list[list[str]] = class_linkage[[parent_col, child_col]].values.tolist()
    # Build a directed graph and a list of all names that have no parent
    graph: dict[str, set] = {name: set() for tup in parent_child_list for name in tup}
    has_parent: dict[str, bool] = {name: False for tup in parent_child_list for name in tup}
    for parent, child in parent_child_list:
        graph[parent].add(child)
        has_parent[child] = True

    # All names that have absolutely no parent:
    roots = [name for name, parents in has_parent.items() if not parents]

    return _traverse({}, graph, roots)


def _traverse(hierarchy: dict, graph: dict, names: list[str]) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def _prettify_generation_order(generation_order: dict, depth: dict | None = None, start=-1) -> dict:
    """Prettifies generation order dictionary for easier consumption."""
    depth = depth or {}
    for key, value in generation_order.items():
        depth[key] = start + 1
        if isinstance(value, dict):
            _prettify_generation_order(value, depth, start=start + 1)
    return OrderedDict(sorted(depth.items(), key=lambda item: item[1]))


def _remove_non_hierarchy_linking(class_linkage: pd.DataFrame) -> pd.DataFrame:
    """Remove linkage which is not creating asset hierarchy."""
    return class_linkage[class_linkage.linking_type == "hierarchy"]


def _remove_higher_occurring_sym_pair(class_linkage: pd.DataFrame, sym_pairs: set[tuple[str, str]]) -> pd.DataFrame:
    """Remove symmetric pair which is higher in occurrence."""
    rows_to_remove = set()
    for source, target in sym_pairs:
        first_sym_property_occurrence = class_linkage[
            (class_linkage.source_class == source) & (class_linkage.target_class == target)
        ].max_occurrence.values[0]
        second_sym_property_occurrence = class_linkage[
            (class_linkage.source_class == target) & (class_linkage.target_class == source)
        ].max_occurrence.values[0]

        if first_sym_property_occurrence is None:
            # this means that source occurrence is unbounded
            index = class_linkage[
                (class_linkage.source_class == source) & (class_linkage.target_class == target)
            ].index.values[0]
        elif (
            second_sym_property_occurrence is None
            or first_sym_property_occurrence <= second_sym_property_occurrence
            and second_sym_property_occurrence > first_sym_property_occurrence
        ):
            # this means that target occurrence is unbounded
            index = class_linkage[
                (class_linkage.source_class == target) & (class_linkage.target_class == source)
            ].index.values[0]
        else:
            index = class_linkage[
                (class_linkage.source_class == source) & (class_linkage.target_class == target)
            ].index.values[0]
        rows_to_remove.add(index)

    return class_linkage.drop(list(rows_to_remove))


def _remove_non_requested_sym_pairs(class_linkage: pd.DataFrame, class_count: dict) -> pd.DataFrame:
    """Remove symmetric pairs which classes are not found in class count."""
    rows_to_remove = set(class_linkage[~(class_linkage["source_class"].isin(set(class_count.keys())))].index.values)
    rows_to_remove |= set(class_linkage[~(class_linkage["target_class"].isin(set(class_count.keys())))].index.values)

    return class_linkage.drop(list(rows_to_remove))


def _generate_mock_data_property_triples(
    instance_ids: list[URIRef], property_: str, namespace: Namespace, value_type: str = "string"
) -> list[tuple[URIRef, URIRef, Literal]]:
    """Generates triples for data properties."""

    python_type = XSD_VALUE_TYPE_MAPPINGS[value_type].python
    triples = []
    for id_ in instance_ids:
        if python_type == int:
            triples.append((id_, URIRef(namespace[property_]), Literal(random.randint(1, 1983))))
        elif python_type == float:
            triples.append((id_, URIRef(namespace[property_]), Literal(numpy.float32(random.uniform(1, 1983)))))
        # generate string
        else:
            triples.append((id_, URIRef(namespace[property_]), Literal(remove_namespace(id_).replace("-", " "))))
    return triples


def _generate_mock_object_property_triples(
    class_: str,
    property_definition: pd.Series,
    class_definitions: dict[str, pd.DataFrame],
    sym_pairs: set[tuple[str, str]],
    instance_ids: dict[str, list[URIRef]],
    namespace: Namespace,
    stop_on_exception: bool,
) -> list[tuple[URIRef, URIRef, URIRef]]:
    """Generates triples for object properties."""
    if property_definition.value_type not in instance_ids:
        msg = f"Class {property_definition.value_type} not found in class count! "
        if stop_on_exception:
            logging.error(msg)
            raise ValueError(msg)
        else:
            msg += (
                f"Skipping creating triples for property {property_definition.name} "
                f"of class {class_} which expects values of this type!"
            )
            logging.warning(msg)
            warnings.warn(msg, stacklevel=2)
            return []

    # Handling symmetric property
    symmetric_property = (
        class_definitions[property_definition.value_type][
            class_definitions[property_definition.value_type].value_type == class_
        ].index.values[0]
        if tuple((class_, property_definition.value_type)) in sym_pairs
        else None
    )

    triples = []

    for i, source in enumerate(instance_ids[class_]):
        target = instance_ids[property_definition.value_type][i % len(instance_ids[property_definition.value_type])]
        triples += [(URIRef(source), URIRef(namespace[str(property_definition.name)]), URIRef(target))]

        if symmetric_property:
            triples += [(URIRef(target), URIRef(namespace[symmetric_property]), URIRef(source))]

    # remove symmetric property from class definition of downstream class
    # to avoid asymmetric linking in mock graph
    if symmetric_property:
        class_definitions[property_definition.value_type].drop(symmetric_property, inplace=True)

    return triples


def _generate_triples_per_class(
    class_: str,
    class_definitions: dict[str, pd.DataFrame],
    sym_pairs: set[tuple[str, str]],
    instance_ids: dict[str, list[URIRef]],
    namespace: Namespace,
    stop_on_exception: bool,
) -> list[Triple]:
    """Generate triples for a given class."""
    triples: list[Triple] = []
    for _, property_definition in class_definitions[class_].iterrows():
        if property_definition.property_type == "DatatypeProperty":
            triples += _generate_mock_data_property_triples(
                instance_ids[class_], str(property_definition.name), namespace, property_definition.value_type
            )

        elif property_definition.property_type == "ObjectProperty":
            triples += _generate_mock_object_property_triples(
                class_, property_definition, class_definitions, sym_pairs, instance_ids, namespace, stop_on_exception
            )

        else:
            logging.error(f"Property type {property_definition.property_type} not supported!")
            raise ValueError(f"Property type {property_definition.property_type} not supported!")

    return triples


def _rules_to_dict(transformation_rules: Rules) -> dict[str, pd.DataFrame]:
    """Represent data model as a dictionary of data frames, where each data frame
    represents properties defined for a given class.

    Args:
        transformation_rules: Transformation rules defining the data model

    Returns:
        Simplified representation of the data model
    """

    data_model: dict[str, pd.DataFrame] = {}

    defined_classes = get_classes_with_properties(transformation_rules)

    for class_ in defined_classes:
        properties = {}
        for property_ in defined_classes[class_]:
            if property_.property_id not in properties:
                properties[property_.property_id] = {
                    "property_type": property_.property_type,
                    "value_type": property_.expected_value_type.suffix,
                    "min_count": property_.min_count,
                    "max_count": property_.max_count,
                }

        data_model[class_] = pd.DataFrame(properties).T
    return data_model



from __future__ import annotations

import itertools
from datetime import datetime
from pathlib import Path
from types import GenericAlias
from typing import Any, ClassVar, Literal, cast, get_args

from openpyxl import Workbook
from openpyxl.cell import MergedCell
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.worksheet.worksheet import Worksheet

from cognite.neat.rules._shared import Rules
from cognite.neat.rules.models._rules import DMSRules, DomainRules, InformationRules
from cognite.neat.rules.models._rules.base import RoleTypes, SchemaCompleteness, SheetEntity

from ._base import BaseExporter


class ExcelExporter(BaseExporter[Workbook]):
    """Export rules to Excel.

    Args:
        styling: The styling to use for the Excel file. Defaults to "default". See below for details
            on the different styles.
        output_role: The role to use for the exported spreadsheet. If provided, the rules will be converted to
            this role formate before being written to excel. If not provided, the role from the rules will be used.
        new_model_id: The new model ID to use for the exported spreadsheet. This is only applicable if the input
            rules have 'is_reference' set. If provided, the model ID will be used to automatically create the
            new metadata sheet in the Excel file.

    The following styles are available:

    - "none":    No styling is applied.
    - "minimal": Column widths are adjusted to fit the content, and the header row(s) is frozen.
    - "default": Minimal + headers are bold, increased size, and colored.
    - "maximal": Default + alternating row colors in the properties sheet for each class in addition to extra
                 blank rows between classes and borders
    """

    Style = Literal["none", "minimal", "default", "maximal"]

    _main_header_by_sheet_name: ClassVar[dict[str, str]] = {
        "Properties": "Definition of Properties per Class",
        "Classes": "Definition of Classes",
        "Views": "Definition of Views",
        "Containers": "Definition of Containers",
    }
    style_options = get_args(Style)

    def __init__(
        self,
        styling: Style = "default",
        output_role: RoleTypes | None = None,
        new_model_id: tuple[str, str, str] | None = None,
    ):
        if styling not in self.style_options:
            raise ValueError(f"Invalid styling: {styling}. Valid options are {self.style_options}")
        self.styling = styling
        self._styling_level = self.style_options.index(styling)
        self.output_role = output_role
        self.new_model_id = new_model_id

    def export_to_file(self, rules: Rules, filepath: Path) -> None:
        """Exports transformation rules to excel file."""
        data = self.export(rules)
        try:
            data.save(filepath)
        finally:
            data.close()
        return None

    def export(self, rules: Rules) -> Workbook:
        rules = self._convert_to_output_role(rules, self.output_role)
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        dumped_rules: dict[str, Any]
        dumped_reference_rules: dict[str, Any] | None = None
        if rules.is_reference:
            # Writes empty reference sheets
            dumped_rules = {
                "Metadata": self._create_metadata_sheet_user_rules(rules),
            }
            dumped_rules["Metadata"]["role"] = (
                self.output_role and self.output_role.value
            ) or rules.metadata.role.value
            dumped_reference_rules = rules.reference_self().model_dump(by_alias=True)
        else:
            dumped_rules = rules.model_dump(by_alias=True)
            if rules.reference:
                dumped_reference_rules = rules.reference.model_dump(by_alias=True)

        self._write_metadata_sheet(workbook, dumped_rules["Metadata"])
        self._write_sheets(workbook, dumped_rules, rules)
        if dumped_reference_rules:
            self._write_sheets(workbook, dumped_reference_rules, rules, is_reference=True)
            self._write_metadata_sheet(workbook, dumped_reference_rules["Metadata"], is_reference=True)

        if self._styling_level > 0:
            self._adjust_column_widths(workbook)

        return workbook

    def _write_sheets(self, workbook: Workbook, dumped_rules: dict[str, Any], rules: Rules, is_reference: bool = False):
        for sheet_name, headers in rules.headers_by_sheet(by_alias=True).items():
            if sheet_name in ("Metadata", "prefixes", "Reference", "is_reference"):
                continue
            if is_reference:
                sheet = workbook.create_sheet(f"Ref{sheet_name}")
            else:
                sheet = workbook.create_sheet(sheet_name)

            # Reorder such that the first column is class + the first field of the subclass
            # of sheet entity. This is to make the properties/classes/views/containers sheet more readable.
            # For example, for the properties these that means class, property, name, description
            # instead of class, name, description, property
            move = len(SheetEntity.model_fields) - 1  # -1 is for the class field
            headers = headers[:1] + headers[move : move + 1] + headers[1:move] + headers[move + 1 :]

            main_header = self._main_header_by_sheet_name[sheet_name]
            sheet.append([main_header] + [""] * (len(headers) - 1))
            sheet.merge_cells(start_row=1, start_column=1, end_row=1, end_column=len(headers))
            sheet.append(headers)

            fill_colors = itertools.cycle(["CADCFC", "FFFFFF"])
            fill_color = next(fill_colors)
            last_class: str | None = None
            item: dict[str, Any]
            for item in dumped_rules.get(sheet_name) or []:
                row = list(item.values())
                class_ = row[0]

                is_properties = sheet_name == "Properties"
                is_new_class = class_ != last_class and last_class is not None
                if self._styling_level > 2 and is_new_class and is_properties:
                    sheet.append([""] * len(headers))
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                    fill_color = next(fill_colors)

                # Need to do the same reordering as for the headers above
                row = row[:1] + row[move : move + 1] + row[1:move] + row[move + 1 :]
                sheet.append(row)
                if self._styling_level > 2 and is_properties:
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                last_class = class_

            if self._styling_level > 0:
                # This freezes all rows above the given row
                sheet.freeze_panes = sheet["A3"]

                sheet["A1"].alignment = Alignment(horizontal="center")

            if self._styling_level > 1:
                # Make the header row bold, larger, and colored
                sheet["A1"].font = Font(bold=True, size=20)
                sheet["A1"].fill = PatternFill(fgColor="FFC000", patternType="solid")
                for cell in sheet["2"]:
                    cell.font = Font(bold=True, size=14)

    def _write_metadata_sheet(self, workbook: Workbook, metadata: dict[str, Any], is_reference: bool = False) -> None:
        # Excel does not support timezone in datetime strings
        if isinstance(metadata.get("created"), datetime):
            metadata["created"] = metadata["created"].replace(tzinfo=None)
        if isinstance(metadata.get("updated"), datetime):
            metadata["updated"] = metadata["updated"].replace(tzinfo=None)

        if is_reference:
            metadata_sheet = workbook.create_sheet("RefMetadata")
        else:
            metadata_sheet = workbook.create_sheet("Metadata")
        for key, value in metadata.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    @classmethod
    def _get_item_class(cls, annotation: GenericAlias) -> type[SheetEntity]:
        if not isinstance(annotation, GenericAlias):
            raise ValueError(f"Expected annotation to be a GenericAlias, but got {type(annotation)}")
        args = get_args(annotation)
        if len(args) != 1:
            raise ValueError(f"Expected annotation to have exactly one argument, but got {len(args)}")
        arg = args[0]
        if not issubclass(arg, SheetEntity):
            raise ValueError(f"Expected annotation to have a BaseModel argument, but got {type(arg)}")
        return arg

    @classmethod
    def _adjust_column_widths(cls, workbook: Workbook) -> None:
        for sheet in workbook:
            sheet = cast(Worksheet, sheet)
            for column_cells in sheet.columns:
                try:
                    max_length = max(len(str(cell.value)) for cell in column_cells if cell.value is not None)
                except ValueError:
                    max_length = 0

                selected_column = column_cells[0]
                if isinstance(selected_column, MergedCell):
                    selected_column = column_cells[1]

                current = sheet.column_dimensions[selected_column.column_letter].width or (max_length + 0.5)
                sheet.column_dimensions[selected_column.column_letter].width = max(current, max_length + 0.5)
        return None

    def _create_metadata_sheet_user_rules(self, rules: Rules) -> dict[str, Any]:
        metadata: dict[str, Any] = {
            field_alias: None for field_alias in rules.metadata.model_dump(by_alias=True).keys()
        }
        if "creator" in metadata:
            metadata["creator"] = "YOUR NAME"

        if isinstance(rules, DomainRules):
            return metadata
        elif isinstance(rules, DMSRules):
            existing_model_id = (rules.metadata.space, rules.metadata.external_id, rules.metadata.version)
        elif isinstance(rules, InformationRules):
            existing_model_id = (rules.metadata.prefix, rules.metadata.name, rules.metadata.version)
        else:
            raise ValueError(f"Unsupported rules type: {type(rules)}")
        existing_metadata = rules.metadata.model_dump(by_alias=True)
        if isinstance(existing_metadata["created"], datetime):
            metadata["created"] = existing_metadata["created"].replace(tzinfo=None)
        if isinstance(existing_metadata["updated"], datetime):
            metadata["updated"] = existing_metadata["updated"].replace(tzinfo=None)
        # Excel does not support timezone in datetime strings
        now_iso = datetime.now().replace(tzinfo=None).isoformat()
        is_info = isinstance(rules, InformationRules)
        is_dms = not is_info
        is_extension = self.new_model_id is not None
        is_solution = is_extension and self.new_model_id != existing_model_id

        if is_solution:
            metadata["prefix" if is_info else "space"] = self.new_model_id[0]  # type: ignore[index]
            metadata["title" if is_info else "externalId"] = self.new_model_id[1]  # type: ignore[index]
            metadata["version"] = self.new_model_id[2]  # type: ignore[index]
        else:
            metadata["prefix" if is_info else "space"] = existing_model_id[0]
            metadata["title" if is_info else "externalId"] = existing_model_id[1]
            metadata["version"] = existing_model_id[2]

        if is_solution and is_info:
            metadata["namespace"] = f"http://purl.org/{self.new_model_id[0]}/"  # type: ignore[index]
        elif is_info:
            metadata["namespace"] = existing_metadata["namespace"]

        if is_solution and is_dms:
            metadata["name"] = self.new_model_id[1]  # type: ignore[index]

        if is_solution:
            metadata["created"] = now_iso
        else:
            metadata["created"] = existing_metadata["created"]

        if is_solution or is_extension:
            metadata["updated"] = now_iso
        else:
            metadata["updated"] = existing_metadata["updated"]

        if is_solution:
            metadata["creator"] = "YOUR NAME"
        else:
            metadata["creator"] = existing_metadata["creator"]

        if not is_solution:
            metadata["description"] = existing_metadata["description"]

        if is_extension:
            metadata["schema"] = SchemaCompleteness.extended.value
        else:
            metadata["schema"] = SchemaCompleteness.complete.value

        metadata["extension"] = "addition"

        return metadata



from cognite.neat.workflows.base import BaseWorkflow
from cognite.neat.workflows.manager import WorkflowManager
from cognite.neat.workflows.model import FlowMessage, WorkflowFullStateReport, WorkflowStepDefinition, WorkflowStepEvent

__all__ = [
    "BaseWorkflow",
    "WorkflowStepDefinition",
    "WorkflowManager",
    "WorkflowStepEvent",
    "WorkflowFullStateReport",
    "FlowMessage",
]



import logging
import time
import warnings
from pathlib import Path
from typing import ClassVar, Literal, cast

from cognite.client import data_modeling as dm

import cognite.neat.graph.extractor._graph_capturing_sheet
from cognite.neat.exceptions import wrangle_warnings
from cognite.neat.rules import exporter
from cognite.neat.rules.exporter._rules2dms import DMSSchemaComponents
from cognite.neat.rules.exporter._rules2graphql import GraphQLSchema
from cognite.neat.rules.exporter._rules2ontology import Ontology
from cognite.neat.utils.utils import generate_exception_report
from cognite.neat.workflows._exceptions import StepNotInitialized
from cognite.neat.workflows.model import FlowMessage, StepExecutionStatus
from cognite.neat.workflows.steps.data_contracts import CogniteClient, DMSSchemaComponentsData, RulesData
from cognite.neat.workflows.steps.step_model import Configurable, Step

__all__ = [
    "ExportDMSSchemaComponentsToYAML",
    "ExportDMSSchemaComponentsToCDF",
    "ExportRulesToGraphQLSchema",
    "ExportRulesToOntology",
    "ExportRulesToSHACL",
    "ExportRulesToGraphCapturingSheet",
    "ExportRulesToExcel",
    "GenerateDMSSchemaComponentsFromRules",
    "DeleteDMSSchemaComponents",
]

CATEGORY = __name__.split(".")[-1].replace("_", " ").title() + " [VERSION 1]"


class GenerateDMSSchemaComponentsFromRules(Step):
    """
    This step generates DMS Schema components, such as data model, views, containers, etc. from Rules.
    """

    description = "This step generates DMS Schema components, such as data model, views, containers, etc. from Rules."
    version = "private-alpha"
    category = CATEGORY

    def run(self, rules: RulesData) -> (FlowMessage, DMSSchemaComponentsData):  # type: ignore[override, syntax]
        data_model = DMSSchemaComponents.from_rules(rules.rules)

        output_text = (
            "DMS Schema Components Generated: "
            f"<li> - {len(data_model.spaces)} spaces</li>"
            f"<li> - {len(data_model.containers)} containers</li>"
            f"<li> - {len(data_model.views)} views</li>"
            f"</ul> which are referred in data model <b><code>{data_model.space}:{data_model.external_id}</code>"
            f"</b>/v=<b><code>{data_model.version}</code></b>"
        )

        # need to store the data model in the step so that it can be used by the next step
        # see GraphQL step

        return FlowMessage(output_text=output_text), DMSSchemaComponentsData(components=data_model)


class ExportDMSSchemaComponentsToYAML(Step):
    """
    This step exports DMS schema components as YAML files
    """

    description = "This step exports DMS schema components as YAML files"
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="storage_dir", value="staging", label="Directory to store DMS schema files"),
        Configurable(
            name="format",
            value="yaml-dump",
            label="Format of the output files",
            options=["yaml-dump", "cognite-toolkit", "all"],
        ),
    ]

    def run(self, data_model_contract: DMSSchemaComponentsData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        staging_dir_str = self.configs["storage_dir"]
        format_ = self.configs["format"]

        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)

        if format_ in ["yaml-dump", "all"]:
            base_file_name = (
                f"{data_model_contract.components.space}-"
                f"{data_model_contract.components.external_id}-"
                f"v{data_model_contract.components.version.strip().replace('.', '_')}"
            )

            _container_file_name = f"{base_file_name}-containers.yaml"
            _data_model_file_name = f"{base_file_name}-data-model.yaml"

            container_full_path = staging_dir / _container_file_name
            data_model_full_path = staging_dir / _data_model_file_name

            data_model = dm.DataModelApply(
                space=data_model_contract.components.space,
                external_id=data_model_contract.components.external_id,
                version=data_model_contract.components.version,
                description=data_model_contract.components.description,
                name=data_model_contract.components.name,
                views=list(data_model_contract.components.views.values()),
            )

            containers = dm.ContainerApplyList(data_model_contract.components.containers.values())

            container_full_path.write_text(containers.dump_yaml())
            data_model_full_path.write_text(data_model.dump_yaml())

            output_text = (
                "<p></p>"
                "DMS Schema exported and can be downloaded here : "
                "<p></p>"
                f'- <a href="/data/{staging_dir_str}/{_data_model_file_name}?{time.time()}" '
                f'target="_blank">{_data_model_file_name}</a>'
                "<p></p>"
                f'- <a href="/data/{staging_dir_str}/{_container_file_name}?{time.time()}" '
                f'target="_blank">{_container_file_name}</a>'
            )

            return FlowMessage(output_text=output_text)
        else:
            return FlowMessage(
                error_text=f"Export format <b><code>{format_}</code></b> not implemented!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )


class ExportDMSSchemaComponentsToCDF(Step):
    """
    This step exports generated DMS Schema components to CDF
    """

    description = "This step exports generated DMS Schema components to CDF."
    version = "private-alpha"
    category = CATEGORY

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="components",
            type="multi_select",
            value="",
            label="Select which DMS schema component(s) to export to CDF",
            options=["space", "container", "view", "data model"],
        ),
        Configurable(
            name="existing_component_handling",
            value="fail",
            label=(
                "How to handle situation when components being exported in CDF already exist."
                "Fail the step if any component already exists, "
                "Skip the component if it already exists, "
                " or Update the component try to update the component."
            ),
            options=["fail", "skip", "update"],
        ),
        Configurable(
            name="multi_space_components_create",
            value="False",
            label=(
                "Whether to create only components belonging to the data model space"
                " (i.e. space define under Metadata sheet of Rules), "
                "or also additionally components outside of the data model space."
            ),
            options=["True", "False"],
        ),
    ]

    def run(self, data_model: DMSSchemaComponentsData, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override, syntax]
        existing_component_handling: str = self.configs["existing_component_handling"]
        multi_space_components_create: bool = self.configs["multi_space_components_create"] == "True"
        components_to_create = {key for key, value in self.complex_configs["components"].items() if value}

        if not components_to_create:
            return FlowMessage(
                error_text="No DMS Schema components selected for upload! Please select minimum one!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        logs, errors = data_model.components.to_cdf(
            cdf_client,
            components_to_create=components_to_create,
            existing_component_handling=cast(Literal["skip"], existing_component_handling),
            multi_space_components_create=multi_space_components_create,
            return_report=True,
        )

        report = "# DMS Schema Components Export to CDF\n\n"
        for component, log in logs.items():
            if log:
                report += f"## {component.upper()}\n" + "\n".join(log) + "\n\n"

        report += "\n\n# ERRORS\n\n"
        for component, log in errors.items():
            if log:
                report += f"## {component.upper()}\n" + "\n".join(log) + "\n\n"

        # report
        report_file = "dms_component_creation_report.txt"
        report_dir = self.data_store_path / Path("staging")
        report_dir.mkdir(parents=True, exist_ok=True)
        report_full_path = report_dir / report_file
        report_full_path.write_text(report)

        output_text = (
            "<p></p>"
            "Download DMS Schema Components export "
            f'<a href="/data/staging/{report_file}?{time.time()}" '
            f'target="_blank">report</a>'
        )

        if any(value for value in errors.values()):
            return FlowMessage(error_text=output_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        else:
            return FlowMessage(output_text=output_text)


class DeleteDMSSchemaComponents(Step):
    """
    This step deletes DMS Schema components
    """

    description = "This step deletes DMS Data model and all underlying containers and views."
    version = "private-alpha"
    category = CATEGORY

    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="components",
            type="multi_select",
            value="",
            label="Select which DMS schema component(s) to delete",
            options=["space", "container", "view", "data model"],
        ),
        Configurable(
            name="multi_space_components_removal",
            value="False",
            label=(
                "False (default) = Only delete components inside the space referred"
                " in Metadata Sheet of Rules"
                r", True = Delete all components referred to in Rules<\p>"
            ),
            options=["True", "False"],
        ),
    ]

    def run(self, data_model: DMSSchemaComponentsData, cdf_client: CogniteClient) -> FlowMessage:  # type: ignore[override, syntax]
        components_to_remove = {key for key, value in self.complex_configs["components"].items() if value}
        multi_space_components_removal: bool = self.configs["multi_space_components_removal"] == "True"
        if not components_to_remove:
            return FlowMessage(
                error_text="No DMS Schema components selected for deletion! Please select minimum one!",
                step_execution_status=StepExecutionStatus.ABORT_AND_FAIL,
            )

        logs, errors = data_model.components.remove(
            cdf_client,
            components_to_remove=components_to_remove,
            multi_space_components_removal=multi_space_components_removal,
            return_report=True,
        )

        report = "# DMS Schema Components Removal from CDF\n\n"
        for component, log in logs.items():
            if log:
                report += f"## {component.upper()}\n" + "\n".join(log) + "\n\n"

        report += "\n\n# ERRORS\n\n"
        for component, log in errors.items():
            if log:
                report += f"## {component.upper()}\n" + "\n".join(log) + "\n\n"

        # report
        report_file = "dms_component_removal_report.txt"
        report_dir = self.data_store_path / Path("staging")
        report_dir.mkdir(parents=True, exist_ok=True)
        report_full_path = report_dir / report_file
        report_full_path.write_text(report)

        output_text = (
            "<p></p>"
            "Download DMS Schema Components removal "
            f'<a href="/data/staging/{report_file}?{time.time()}" '
            f'target="_blank">report</a>'
        )

        if any(value for value in errors.values()):
            return FlowMessage(error_text=output_text, step_execution_status=StepExecutionStatus.ABORT_AND_FAIL)
        else:
            return FlowMessage(output_text=output_text)


class ExportRulesToGraphQLSchema(Step):
    """
    This step generates GraphQL schema from data model defined in transformation rules
    """

    description = "This step generates GraphQL schema from data model defined in transformation rules."
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="file_name",
            value="",
            label=(
                "Name of the GraphQL schema file it must have .graphql extension,"
                " if empty defaults to form `prefix-version.graphql`"
            ),
        ),
        Configurable(name="storage_dir", value="staging", label="Directory to store GraphQL schema file"),
    ]

    def run(self, transformation_rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        data_model_gql = GraphQLSchema.from_rules(transformation_rules.rules, verbose=True).schema

        default_name = (
            f"{transformation_rules.rules.metadata.prefix}-"
            f"v{transformation_rules.rules.metadata.version.strip().replace('.', '_')}"
            ".graphql"
        )

        schema_name = self.configs["file_name"] or default_name

        staging_dir_str = self.configs["storage_dir"]
        staging_dir = self.data_store_path / Path(staging_dir_str)
        staging_dir.mkdir(parents=True, exist_ok=True)
        fdm_model_full_path = staging_dir / schema_name

        fdm_model_full_path.write_text(data_model_gql)

        output_text = (
            "<p></p>"
            "GraphQL Schema generated and can be downloaded here : "
            f'<a href="/data/{staging_dir_str}/{schema_name}?{time.time()}" '
            f'target="_blank">{schema_name}</a>'
        )

        return FlowMessage(output_text=output_text)


class ExportRulesToOntology(Step):
    """
    This step exports Rules to OWL ontology
    """

    description = "This step exports Rules to OWL ontology"
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="ontology_file_path",
            value="staging/ontology.ttl",
            label=(
                "Relative path for the ontology file storage, "
                "must end with .ttl ! Will be auto-created if not provided !"
            ),
        )
    ]

    def run(self, rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        # ontology file
        default_path = self.data_store_path / Path(
            f"{rules.rules.metadata.prefix}-"
            f"v{rules.rules.metadata.version.strip().replace('.', '_')}"
            "-ontology.ttl"
        )

        if not self.configs["ontology_file_path"]:
            storage_path = default_path
        else:
            storage_path = self.data_store_path / Path(self.configs["ontology_file_path"])

        storage_path.parent.mkdir(parents=True, exist_ok=True)
        report_file_path = storage_path.parent / f"report_{storage_path.stem}.txt"

        with warnings.catch_warnings(record=True) as validation_warnings:
            ontology = Ontology.from_rules(rules=rules.rules)

        storage_path.write_text(ontology.ontology)
        report_file_path.write_text(generate_exception_report(wrangle_warnings(validation_warnings), "Warnings"))

        relative_ontology_file_path = str(storage_path).split("/data/")[1]

        output_text = (
            "<p></p>"
            "Rules exported to ontology can be downloaded here : "
            f'<a href="/data/{relative_ontology_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)


class ExportRulesToSHACL(Step):
    """
    This step exports Rules to SHACL
    """

    description = "This step exports Rules to SHACL"
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="shacl_file_path",
            value="staging/shacl.ttl",
            label=(
                "Relative path for the SHACL file storage, "
                "must end with .ttl ! Will be auto-created if not provided !"
            ),
        )
    ]

    def run(self, rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)

        # ontology file
        default_path = self.data_store_path / Path(
            f"{rules.rules.metadata.prefix}-" f"v{rules.rules.metadata.version.strip().replace('.', '_')}" "-shacl.ttl"
        )

        if not self.configs["shacl_file_path"]:
            storage_path = default_path
        else:
            storage_path = self.data_store_path / Path(self.configs["shacl_file_path"])
        report_file_path = storage_path.parent / f"report_{storage_path.stem}.txt"

        with warnings.catch_warnings(record=True) as validation_warnings:
            ontology = Ontology.from_rules(rules=rules.rules)

        storage_path.parent.mkdir(parents=True, exist_ok=True)
        storage_path.write_text(ontology.constraints)
        report_file_path.write_text(generate_exception_report(wrangle_warnings(validation_warnings), "Warnings"))

        relative_shacl_file_path = str(storage_path).split("/data/")[1]

        output_text = (
            "<p></p>"
            "Rules exported to ontology can be downloaded here : "
            f'<a href="/data/{relative_shacl_file_path}?{time.time()}" '
            f'target="_blank">{storage_path.stem}.ttl</a>'
        )

        return FlowMessage(output_text=output_text)


class ExportRulesToGraphCapturingSheet(Step):
    """
    This step generates graph capturing sheet
    """

    description = "This step generates graph capturing sheet"
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(name="file_name", value="graph_capture_sheet.xlsx", label="File name of the data capture sheet"),
        Configurable(name="auto_identifier_type", value="index-based", label="Type of automatic identifier"),
        Configurable(name="storage_dir", value="staging", label="Directory to store data capture sheets"),
    ]

    def run(self, rules: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        if self.configs is None or self.data_store_path is None:
            raise StepNotInitialized(type(self).__name__)
        logging.info("Generate graph capture sheet")
        sheet_name = self.configs["file_name"]
        auto_identifier_type = self.configs["auto_identifier_type"]
        staging_dir_str = self.configs["storage_dir"]
        logging.info(f"Auto identifier type {auto_identifier_type}")
        staging_dir = self.data_store_path / Path(staging_dir_str)

        staging_dir.mkdir(parents=True, exist_ok=True)

        data_capture_sheet_path = staging_dir / sheet_name

        cognite.neat.graph.extractor._graph_capturing_sheet.rules2graph_capturing_sheet(
            rules.rules, data_capture_sheet_path, auto_identifier_type=auto_identifier_type
        )

        output_text = (
            "Data capture sheet generated and can be downloaded here : "
            f'<a href="/data/{staging_dir_str}/{sheet_name}?{time.time()}" target="_blank">'
            f"{sheet_name}</a>"
        )
        return FlowMessage(output_text=output_text)


class ExportRulesToExcel(Step):
    description = "This step export Rules to Excel representation"
    version = "private-alpha"
    category = CATEGORY
    configurables: ClassVar[list[Configurable]] = [
        Configurable(
            name="output_file_path", value="rules/custom-rules.xlsx", label="File path to the generated Excel file"
        )
    ]

    def run(self, rules_data: RulesData) -> FlowMessage:  # type: ignore[override, syntax]
        full_path = Path(self.data_store_path) / Path(self.configs["output_file_path"])
        exporter.ExcelExporter.from_rules(rules=rules_data.rules).export_to_file(filepath=full_path)
        return FlowMessage(output_text="Generated Excel file from rules")



import importlib
import inspect
import logging
import os
import sys
import types
from pathlib import Path
from typing import Any

from pydantic import BaseModel

import cognite.neat.workflows.steps.lib
import cognite.neat.workflows.steps.lib.v1
from cognite.neat.app.monitoring.metrics import NeatMetricsCollector
from cognite.neat.exceptions import InvalidWorkFlowError
from cognite.neat.workflows._exceptions import ConfigurationNotSet
from cognite.neat.workflows.model import FlowMessage, WorkflowConfigs
from cognite.neat.workflows.steps.step_model import Configurable, DataContract, Step


class StepMetadata(BaseModel):
    name: str
    description: str = ""
    category: str = "default"  # defines the category the step belongs to (e.g. "data", "model", "test")
    scope: str = "global"  # defines the scope of the step (e.g. "global", local to a specific workflow)
    input: list[str]
    output: list[str]
    configurables: list[Configurable] = []
    version: str = "1.0.0"  # version of the step. All alpha versions considered as experimental
    docs_url: str = ""  # url to the extended documentation of the step
    source: str = (
        "cognite"  # source of the step , can be source identifier or url , for instance github url for instance.
    )


class StepsRegistry:
    def __init__(self, data_store_path: Path | None = None):
        self._step_classes: list[type[Step]] = []
        if data_store_path:
            self.user_steps_path: Path | None = Path(data_store_path) / "steps"
        else:
            self.user_steps_path = None
        self.data_store_path: str = str(data_store_path)

    def load_step_classes(self):
        if self._step_classes:
            # classes already loaded - no need to reload
            return
        for name, step_cls in inspect.getmembers(cognite.neat.workflows.steps.lib):
            if inspect.isclass(step_cls):
                logging.info(f"Loading NEAT step {name}")
                self._step_classes.append(step_cls)
        for name, step_cls in inspect.getmembers(cognite.neat.workflows.steps.lib.v1):
            if inspect.isclass(step_cls):
                logging.info(f"Loading NEAT step {name}")
                self._step_classes.append(step_cls)
        sys.path.append(str(Path(self.data_store_path) / "workflows"))
        try:
            if self.user_steps_path:
                sys.path.append(str(self.user_steps_path))
                self.load_custom_step_classes(self.user_steps_path, scope="user_global")
        except Exception as e:
            logging.info(f"No user defined modules provided in {self.user_steps_path}. Error: {e}")

    def load_workflow_step_classes(self, workflow_name: str):
        if not self.data_store_path:
            raise ConfigurationNotSet("data_store_path")
        workflow_steps_path = Path(self.data_store_path) / "workflows" / workflow_name
        if workflow_steps_path.exists():
            self.load_custom_step_classes(workflow_steps_path, scope="workflow")

    def load_custom_step_classes(self, custom_steps_path: Path, scope: str = "global"):
        for step_module_name in os.listdir(custom_steps_path):
            step_module_path = custom_steps_path / Path(step_module_name)
            if step_module_name.startswith("__") or (
                step_module_path.is_file() and not step_module_name.endswith(".py")
            ):
                continue
            if scope == "user_global":
                full_module_name = step_module_name.replace(".py", "")
            else:
                full_module_name = custom_steps_path.name + "." + step_module_name.replace(".py", "")
            logging.info(f"Loading user defined step module {full_module_name}")
            if full_module_name in sys.modules:
                logging.info(f"Reloading existing workflow module {full_module_name}")
                steps_module = importlib.reload(sys.modules[full_module_name])
            else:
                steps_module = importlib.import_module(full_module_name)
                logging.info(f"Loading user defined step from {steps_module}")
            for name, step_cls in inspect.getmembers(steps_module):
                base_class = getattr(step_cls, "__bases__", None)
                if (
                    name.startswith("__")
                    or base_class is None
                    or len(base_class) == 0
                    or base_class[0].__name__ != "Step"
                ):
                    continue
                logging.info(f"Loading user defined step {name} from {steps_module}")
                if inspect.isclass(step_cls):
                    step_cls.scope = scope
                    is_new_class = True
                    for i, step in enumerate(self._step_classes):
                        if step.__name__ == step_cls.__name__:
                            logging.info(f"Reloading user defined step {name} class {step_cls.__name__}")
                            self._step_classes[i] = step_cls
                            is_new_class = False
                    if is_new_class:
                        logging.info(f"Loading NEW user defined step {name} class {step_cls.__name__}")
                        self._step_classes.append(step_cls)

    def run_step(
        self,
        step_name: str,
        flow_context: dict[str, DataContract],
        step_configs: dict[str, Any],
        metrics: NeatMetricsCollector | None = None,
        workflow_configs: WorkflowConfigs | None = None,
        workflow_id: str = "",
        workflow_run_id: str = "",
        step_complex_configs: dict[str, Any] | None = None,
    ) -> DataContract | tuple[FlowMessage, DataContract] | FlowMessage:
        if step_complex_configs is None:
            step_complex_configs = {}
        for step_cls in self._step_classes:
            if step_cls.__name__ == step_name:
                step_obj: Step = step_cls(Path(self.data_store_path))
                step_obj.configure(step_configs, step_complex_configs)
                step_obj.set_flow_context(flow_context)
                step_obj.set_metrics(metrics)
                step_obj.set_workflow_configs(workflow_configs)
                step_obj.set_workflow_metadata(workflow_id, workflow_run_id)
                signature = inspect.signature(step_obj.run)
                parameters = signature.parameters
                is_valid = True
                input_data = []
                missing_data = []
                for parameter in parameters.values():
                    try:
                        if parameter.annotation is FlowMessage:
                            input_data.append(flow_context["FlowMessage"])
                        else:
                            if isinstance(parameter.annotation, types.UnionType):
                                for param in parameter.annotation.__args__:
                                    if (
                                        param.__name__ != "_empty"
                                        and param.__name__ in flow_context
                                        and flow_context[param.__name__] is not None
                                    ):
                                        input_data.append(flow_context[param.__name__])
                                        break  # Only one variable can be used as input
                            else:
                                input_data.append(flow_context[parameter.annotation.__name__])
                    except KeyError:
                        is_valid = False
                        logging.error(f"Missing data for step {step_name} parameter {parameter.name}")
                        missing_data.append(parameter.annotation.__name__)
                        continue
                if not is_valid:
                    raise InvalidWorkFlowError(step_name, missing_data)
                return step_obj.run(*input_data)
        raise InvalidWorkFlowError(step_name, [])

    def get_list_of_steps(self) -> list[StepMetadata]:
        steps: list[StepMetadata] = []
        for step_cls in self._step_classes:
            try:
                signature = inspect.signature(step_cls.run)
                parameters = signature.parameters
                input_data = []
                output_data = []
                for parameter in parameters.values():
                    if isinstance(parameter.annotation, types.UnionType):
                        for param in parameter.annotation.__args__:
                            if param.__name__ != "_empty":
                                input_data.append(param.__name__)
                    elif parameter.annotation.__name__ != "_empty":
                        input_data.append(parameter.annotation.__name__)
                return_annotation = signature.return_annotation
                if return_annotation:
                    if isinstance(return_annotation, types.UnionType):
                        for annotation in return_annotation.__args__:
                            output_data.append(annotation.__name__)
                    elif isinstance(return_annotation, tuple):
                        for annotation in return_annotation:
                            if isinstance(annotation, types.UnionType):
                                for annotation_l2 in annotation.__args__:
                                    output_data.append(annotation_l2.__name__)
                            else:
                                output_data.append(annotation.__name__)
                    else:
                        output_data.append(return_annotation.__name__)
                steps.append(
                    StepMetadata(
                        name=step_cls.__name__,
                        scope=step_cls.scope,
                        input=input_data,
                        description=step_cls.description,
                        category=step_cls.category,
                        output=output_data,
                        configurables=step_cls.configurables,
                        version=step_cls.version,
                        source=step_cls.source,
                        docs_url=step_cls.docs_url,
                    )
                )
            except AttributeError as e:
                logging.error(
                    f"Step {type(step_cls).__name__} does not have a run method or types can't be infered.Error: {e}"
                )

        return steps



import sys
from abc import ABC, abstractmethod
from collections import UserList
from collections.abc import Sequence
from dataclasses import dataclass
from functools import total_ordering
from typing import Any, ClassVar
from warnings import WarningMessage

import pandas as pd
from pydantic_core import ErrorDetails

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup
else:
    pass

__all__ = [
    "ValidationIssue",
    "NeatValidationError",
    "DefaultPydanticError",
    "ValidationWarning",
    "DefaultWarning",
    "IssueList",
    "MultiValueError",
]


@total_ordering
@dataclass(frozen=True)
class ValidationIssue(ABC):
    description: ClassVar[str]
    fix: ClassVar[str]

    def message(self) -> str:
        """Return a human-readable message for the issue.

        This is the default implementation, which returns the description.
        It is recommended to override this method in subclasses with a more
        specific message.
        """
        return self.description

    @abstractmethod
    def dump(self) -> dict[str, Any]:
        """Return a dictionary representation of the issue."""
        raise NotImplementedError()

    def __lt__(self, other: "ValidationIssue") -> bool:
        if not isinstance(other, ValidationIssue):
            return NotImplemented
        return (type(self).__name__, self.message()) < (type(other).__name__, other.message())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ValidationIssue):
            return NotImplemented
        return (type(self).__name__, self.message()) == (type(other).__name__, other.message())


@dataclass(frozen=True)
class NeatValidationError(ValidationIssue, ABC):
    def dump(self) -> dict[str, Any]:
        return {"error": type(self).__name__}

    @classmethod
    def from_pydantic_errors(cls, errors: list[ErrorDetails], **kwargs) -> "list[NeatValidationError]":
        """Convert a list of pydantic errors to a list of Error instances.

        This is intended to be overridden in subclasses to handle specific error types.
        """
        all_errors = []
        for error in errors:
            if isinstance(ctx := error.get("ctx"), dict) and isinstance(
                multi_error := ctx.get("error"), MultiValueError
            ):
                all_errors.extend(multi_error.errors)
            else:
                all_errors.append(DefaultPydanticError.from_pydantic_error(error))
        return all_errors


@dataclass(frozen=True)
class DefaultPydanticError(NeatValidationError):
    type: str
    loc: tuple[int | str, ...]
    msg: str
    input: Any
    ctx: dict[str, Any] | None

    @classmethod
    def from_pydantic_error(cls, error: ErrorDetails) -> "NeatValidationError":
        return cls(
            type=error["type"],
            loc=error["loc"],
            msg=error["msg"],
            input=error.get("input"),
            ctx=error.get("ctx"),
        )

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["type"] = self.type
        output["loc"] = self.loc
        output["msg"] = self.msg
        output["input"] = self.input
        output["ctx"] = self.ctx
        return output

    def message(self) -> str:
        if self.loc and len(self.loc) == 1:
            return f"{self.loc[0]} sheet: {self.msg}"
        elif self.loc and len(self.loc) == 2:
            return f"{self.loc[0]} sheet field/column <{self.loc[1]}>: {self.msg}"
        else:
            return self.msg


@dataclass(frozen=True)
class ValidationWarning(ValidationIssue, ABC, UserWarning):
    def dump(self) -> dict[str, Any]:
        return {"warning": type(self).__name__}

    @classmethod
    def from_warning(cls, warning: WarningMessage) -> "ValidationWarning":
        return DefaultWarning.from_warning_message(warning)


@dataclass(frozen=True)
class DefaultWarning(ValidationWarning):
    description = "A warning was raised during validation."
    fix = "No fix is available."

    warning: str | Warning
    category: type[Warning]
    source: str | None = None

    def dump(self) -> dict[str, Any]:
        output = super().dump()
        output["msg"] = str(self.warning)
        output["category"] = self.category.__name__
        output["source"] = self.source
        return output

    @classmethod
    def from_warning_message(cls, warning: WarningMessage) -> "ValidationWarning":
        if isinstance(warning.message, ValidationWarning):
            return warning.message

        return cls(
            warning=warning.message,
            category=warning.category,
            source=warning.source,
        )

    def message(self) -> str:
        return str(self.warning)


class IssueList(UserList[ValidationIssue]):
    def __init__(self, issues: Sequence[ValidationIssue] | None = None, title: str | None = None):
        super().__init__(issues or [])
        self.title = title

    @property
    def errors(self) -> "IssueList":
        return IssueList([issue for issue in self if isinstance(issue, NeatValidationError)])

    @property
    def has_errors(self) -> bool:
        return any(isinstance(issue, NeatValidationError) for issue in self)

    @property
    def warnings(self) -> "IssueList":
        return IssueList([issue for issue in self if isinstance(issue, ValidationWarning)])

    def as_errors(self) -> ExceptionGroup:
        return ExceptionGroup(
            "Validation failed",
            [ValueError(issue.message()) for issue in self if isinstance(issue, NeatValidationError)],
        )

    def to_pandas(self) -> pd.DataFrame:
        return pd.DataFrame([issue.dump() for issue in self])

    def _repr_html_(self) -> str | None:
        return self.to_pandas()._repr_html_()  # type: ignore[operator]


class MultiValueError(ValueError):
    """This is a container for multiple errors.

    It is used in the pydantic field_validator/model_validator to collect multiple errors, which
    can then be caught in a try-except block and returned as an IssueList.

    """

    def __init__(self, errors: Sequence[NeatValidationError]):
        self.errors = list(errors)



import getpass
from abc import ABC, abstractmethod
from datetime import datetime

import pandas as pd
from pydantic_core import ErrorDetails

from cognite.neat.rules.models.raw_rules import RawRules
from cognite.neat.rules.models.rules import Rules


class BaseImporter(ABC):
    """
    BaseImporter class which all importers inherit from.
    """

    @abstractmethod
    def __init__(self):
        ...

    @abstractmethod
    def to_tables(self) -> dict[str, pd.DataFrame]:
        """Creates raw tables from the data."""
        raise NotImplementedError

    def to_raw_rules(self) -> RawRules:
        """Creates `RawRules` object from the data."""

        tables = self.to_tables()

        return RawRules.from_tables(tables=tables, importer_type=self.__class__.__name__)

    def to_rules(
        self,
        return_report: bool = False,
        skip_validation: bool = False,
        validators_to_skip: set[str] | None = None,
    ) -> tuple[Rules | None, list[ErrorDetails] | None, list | None] | Rules:
        """
        Creates `Rules` object from the data.

        Args:
            return_report: To return validation report. Defaults to False.
            skip_validation: Bypasses Rules validation. Defaults to False.
            validators_to_skip: List of validators to skip. Defaults to None.

        Returns:
            Instance of `Rules`, which can be validated, not validated based on
            `skip_validation` flag, or partially validated if `validators_to_skip` is set,
            and optional list of errors and warnings if
            `return_report` is set to True.

        !!! Note "Skip Validation
            `skip_validation` flag should be only used for purpose when `Rules` object
            is exported to an Excel file. Do not use this flag for any other purpose!
        """

        raw_rules = self.to_raw_rules()

        return raw_rules.to_rules(return_report, skip_validation, validators_to_skip)

    def _default_metadata(self):
        return {
            "prefix": "neat",
            "version": "0.1.0",
            "title": "Neat Imported Data Model",
            "created": datetime.now().replace(microsecond=0).isoformat(),
            "creator": getpass.getuser(),
            "description": f"Imported using {type(self).__name__}",
        }



from ._base import BaseImporter
from ._dict2rules import ArbitraryDictImporter
from ._dms2rules import DMSImporter
from ._graph2rules import GraphImporter
from ._json2rules import ArbitraryJSONImporter
from ._owl2rules import OWLImporter
from ._spreadsheet2rules import ExcelImporter, GoogleSheetImporter
from ._xsd2rules import XSDImporter
from ._yaml2rules import ArbitraryYAMLImporter

__all__ = [
    "BaseImporter",
    "ArbitraryDictImporter",
    "ArbitraryJSONImporter",
    "ArbitraryYAMLImporter",
    "DMSImporter",
    "OWLImporter",
    "XSDImporter",
    "GraphImporter",
    "ExcelImporter",
    "GoogleSheetImporter",
]



from cognite.neat.graph.loader.core.labels import upload_labels

from ._asset_loader import AssetLoader
from ._base import BaseLoader, CogniteLoader
from .core.rdf_to_assets import categorize_assets, rdf2assets, upload_assets
from .core.rdf_to_relationships import categorize_relationships, rdf2relationships, upload_relationships
from .rdf_to_dms import DMSLoader, upload_edges, upload_nodes

__all__ = [
    "BaseLoader",
    "CogniteLoader",
    "AssetLoader",
    "rdf2relationships",
    "categorize_assets",
    "upload_assets",
    "rdf2assets",
    "categorize_relationships",
    "upload_relationships",
    "upload_labels",
    "upload_nodes",
    "upload_edges",
    "DMSLoader",
]
